#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import unittest

import numpy as np

from cil.framework import AcquisitionGeometry, ImageGeometry
from cil.framework.labels import FillType, AngleUnit, AcquisitionType, ImageDimension, AcquisitionDimension, Backend


class Test_Lables(unittest.TestCase):
    def test_labels_strenum(self):
        for item in (AngleUnit.DEGREE, "DEGREE", "degree"):
            out = AngleUnit(item)
            self.assertEqual(out, AngleUnit.DEGREE)
            self.assertTrue(isinstance(out, AngleUnit))
        for item in ("bad_str", FillType.RANDOM):
            with self.assertRaises(ValueError):
                AngleUnit(item)

    def test_labels_strenum_eq(self):
        for i in (AngleUnit.RADIAN, "RADIAN", "radian"):
            self.assertEqual(AngleUnit.RADIAN, i)
            self.assertEqual(i, AngleUnit.RADIAN)
        for i in ("DEGREE", AngleUnit.DEGREE, AngleUnit):
            self.assertNotEqual(AngleUnit.RADIAN, i)

    def test_labels_contains(self):
        for i in ("RADIAN", "degree", AngleUnit.RADIAN, AngleUnit.DEGREE):
            self.assertIn(i, AngleUnit)
        for i in ("bad_str", AngleUnit):
            self.assertNotIn(i, AngleUnit)

    def test_backends(self):
        for i in ('ASTRA', 'CIL', 'TIGRE'):
            self.assertIn(i, Backend)
            self.assertIn(i.lower(), Backend)
            self.assertIn(getattr(Backend, i), Backend)

    def test_fill_types(self):
        for i in ('RANDOM', 'RANDOM_INT', 'RANDOM_DEPRECATED', 'RANDOM_INT_DEPRECATED'):
            self.assertIn(i, FillType)
            self.assertIn(i.lower(), FillType)
            self.assertIn(getattr(FillType, i), FillType)

    def test_units_angles(self):
        for i in ('DEGREE', 'RADIAN'):
            self.assertIn(i, AngleUnit)
            self.assertIn(i.lower(), AngleUnit)
            self.assertIn(getattr(AngleUnit, i), AngleUnit)

    def test_acquisition_type(self):
        for i in ('PARALLEL', 'CONE', 'DIM2', 'DIM3', 'CONE_FLEX'):
            self.assertIn(i, AcquisitionType)
            self.assertIn(i.lower(), AcquisitionType)
            self.assertIn(getattr(AcquisitionType, i), AcquisitionType)
        combo = AcquisitionType.DIM2 | AcquisitionType.CONE
        for i in (AcquisitionType.DIM2, AcquisitionType.CONE):
            self.assertIn(i, combo)
        for i in (AcquisitionType.DIM3, AcquisitionType.PARALLEL):
            self.assertNotIn(i, combo)
        for i in ('2D', 'DIM2', AcquisitionType.DIM2):
            self.assertEqual(combo.dimension, i)
        for i in ('CONE', 'cone', AcquisitionType.CONE):
            self.assertEqual(combo.geometry, i)

    def test_image_dimension_labels(self):
        for i in ('CHANNEL', 'VERTICAL', 'HORIZONTAL_X', 'HORIZONTAL_Y'):
            self.assertIn(i, ImageDimension)
            self.assertIn(i.lower(), ImageDimension)
            self.assertIn(getattr(ImageDimension, i), ImageDimension)

    def test_image_dimension_labels_default_order(self):
        order_gold = ImageDimension.CHANNEL, 'VERTICAL', 'horizontal_y', 'HORIZONTAL_X'
        for i in ('CIL', 'TIGRE', 'ASTRA'):
            self.assertSequenceEqual(ImageDimension.get_order_for_engine(i), order_gold)

        with self.assertRaises((KeyError, ValueError)):
            AcquisitionDimension.get_order_for_engine("bad_engine")

    def test_image_dimension_labels_get_order(self):
        ig = ImageGeometry(4, 8, 1, channels=2)
        ig.set_labels(['channel', 'horizontal_y', 'horizontal_x'])

        # for 2D all engines have the same order
        order_gold = ImageDimension.CHANNEL, 'HORIZONTAL_Y', 'horizontal_x'
        self.assertSequenceEqual(ImageDimension.get_order_for_engine('cil', ig), order_gold)
        self.assertSequenceEqual(ImageDimension.get_order_for_engine('tigre', ig), order_gold)
        self.assertSequenceEqual(ImageDimension.get_order_for_engine('astra', ig), order_gold)

    def test_image_dimension_labels_check_order(self):
        ig = ImageGeometry(4, 8, 1, channels=2)
        ig.set_labels(['horizontal_x', 'horizontal_y', 'channel'])

        for i in ('cil', 'tigre', 'astra'):
            with self.assertRaises(ValueError):
                ImageDimension.check_order_for_engine(i, ig)

        ig.set_labels(['channel', 'horizontal_y', 'horizontal_x'])
        self.assertTrue(ImageDimension.check_order_for_engine("cil", ig))
        self.assertTrue(ImageDimension.check_order_for_engine("tigre", ig))
        self.assertTrue(ImageDimension.check_order_for_engine("astra", ig))

    def test_acquisition_dimension_labels(self):
        for i in ('CHANNEL', 'ANGLE', 'VERTICAL', 'HORIZONTAL'):
            self.assertIn(i, AcquisitionDimension)
            self.assertIn(i.lower(), AcquisitionDimension)
            self.assertIn(getattr(AcquisitionDimension, i), AcquisitionDimension)

    def test_acquisition_dimension_labels_default_order(self):
        gold = AcquisitionDimension.CHANNEL, 'ANGLE', 'vertical', 'HORIZONTAL'
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine('CIL'), gold)
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine(Backend.TIGRE), gold)
        gold = 'CHANNEL', 'VERTICAL', 'ANGLE', 'HORIZONTAL'
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine('astra'), gold)

        with self.assertRaises((KeyError, ValueError)):
            AcquisitionDimension.get_order_for_engine("bad_engine")

    def test_acquisition_dimension_labels_get_order(self):
        ag = AcquisitionGeometry.create_Parallel2D()\
            .set_angles(np.arange(0,16 , 1), angle_unit="degree")\
            .set_panel(4)\
            .set_channels(8)\
            .set_labels(['angle', 'horizontal', 'channel'])

        # for 2D all engines have the same order
        order_gold = AcquisitionDimension.CHANNEL, 'ANGLE', 'horizontal'
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine('CIL', ag), order_gold)
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine('TIGRE', ag), order_gold)
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine('ASTRA', ag), order_gold)

        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_angles(np.arange(0,16 , 1), angle_unit="degree")\
            .set_panel((4,2))\
            .set_labels(['angle', 'horizontal', 'vertical'])

        order_gold = AcquisitionDimension.ANGLE, 'VERTICAL', 'horizontal'
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine("cil", ag), order_gold)
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine("tigre", ag), order_gold)
        order_gold = AcquisitionDimension.VERTICAL, 'ANGLE', 'horizontal'
        self.assertSequenceEqual(AcquisitionDimension.get_order_for_engine("astra", ag), order_gold)

    def test_acquisition_dimension_labels_check_order(self):
        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_angles(np.arange(0,16 , 1), angle_unit="degree")\
            .set_panel((8,4))\
            .set_channels(2)\
            .set_labels(['angle', 'horizontal', 'channel', 'vertical'])

        for i in ('cil', 'tigre', 'astra'):
            with self.assertRaises(ValueError):
                AcquisitionDimension.check_order_for_engine(i, ag)

        ag.set_labels(['channel', 'angle', 'vertical', 'horizontal'])
        self.assertTrue(AcquisitionDimension.check_order_for_engine("cil", ag))
        self.assertTrue(AcquisitionDimension.check_order_for_engine("tigre", ag))

        ag.set_labels(['channel', 'vertical', 'angle', 'horizontal'])
        self.assertTrue(AcquisitionDimension.check_order_for_engine("astra", ag))




from utils import initialise_tests
from cil.framework import ImageGeometry
from cil.optimisation.operators import GradientOperator, AdjointOperator, MatrixOperator
import numpy as np 


from testclass import CCPiTestClass


initialise_tests()

class TestAdjointOperator(CCPiTestClass):
    
    def setUp(self):
        self.ig_real = ImageGeometry(3,4)
    
    def tearDown(self):
        pass   
    
    def test_direct_adjoint_gradient_operator(self):

        G = GradientOperator(self.ig_real)
        x = G.domain_geometry().allocate("random", seed=10)
        y = G.range_geometry().allocate("random", seed=10)
        G_adjoint = AdjointOperator(G)

        res1 = G.adjoint(y)
        res2 = G_adjoint.direct(y)
        np.testing.assert_allclose(res1.array, res2.array, atol=1e-6)

        res3 = G.direct(x)
        res4 = G_adjoint.adjoint(x)
        np.testing.assert_allclose(res3[0].array, res4[0].array, atol=1e-6)
        np.testing.assert_allclose(res3[1].array, res4[1].array, atol=1e-6)
        
    def test_direct_adjoint_matrix_operator(self):
        
        np.random.seed(10)
        n = 3  
        m = 5 
        Anp = np.random.normal(0,1, (m, n)).astype('float32')
        
        Amat = MatrixOperator(Anp)
        Amat_tr = MatrixOperator(Anp.T)

        x = Amat.domain_geometry().allocate("random", seed=10)
        y = Amat.range_geometry().allocate("random", seed=11)

        res1 = Amat.adjoint(y)
        res2 = Amat_tr.direct(y)
        np.testing.assert_allclose(res1.array, res2.array, atol=1e-6)

        res3 = Amat.direct(x)
        res4 = Amat_tr.adjoint(x)
        np.testing.assert_allclose(res3.array, res4.array, atol=1e-6) 

        self.assertTrue(Amat.dot_test(Amat), True)  
        self.assertTrue(Amat_tr.dot_test(Amat_tr), True)

        # <Ax, y> = <x, A^T y>
        lhs = res3.dot(y)
        rhs_a = x.dot(res2)
        rhs_b = x.dot(Amat.adjoint(y))
        np.testing.assert_allclose(lhs, rhs_a, atol=1e-6) 
        np.testing.assert_allclose(lhs, rhs_b, atol=1e-6) 

    def test_direct_adjoint_matrix_operator_complex(self):
        
        np.random.seed(10)
        n = 3  
        m = 2
        Anp = np.random.uniform(0,1, (n, m)) - 1.j*np.random.uniform(0,1, (n, m)) + 3.j*np.random.uniform(0,1, (n, m))
        
        Amat = MatrixOperator(Anp)
        Amat_tr = AdjointOperator(Amat)

        x = Amat.domain_geometry().allocate("random", dtype="complex", seed=10)
        y = Amat.range_geometry().allocate("random", dtype="complex", seed=11)


        # <Ax,y> = <x, A^* y> using numpy arrays
        # convention is to conjugate the second vector, see `dot` method in `framework`
        lhs_np = np.dot(Amat.direct(x).array,y.array.conjugate())
        rhs_np = np.dot(x.array, Amat.adjoint(y).conjugate().array)
        np.testing.assert_allclose(lhs_np, rhs_np, atol=1e-6) 

        # using `dot` method from DataContainer
        lhs = Amat.direct(x).dot(y) # conjugate of y is applied in the dot method
        rhs_a = x.dot(Amat.adjoint(y)) # conjugate of y is applied in the dot method
        rhs_b = x.dot(Amat_tr.direct(y)) 
        np.testing.assert_allclose(lhs, rhs_a, atol=1e-6)  
        np.testing.assert_allclose(lhs, rhs_b, atol=1e-6)         


    def test_adjoint_operator_and_dot(self):
        ig = ImageGeometry(3,4,  dtype="complex")
        G = GradientOperator(ig)
        div = AdjointOperator(G)

        x = G.domain.allocate("random_int")
        y = G.range.allocate("random_int")
        res1 = G.direct(x).dot(y)
        res2 = x.dot(div.direct(y))
        self.assertAlmostEqual(res1,res2)


#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
from enum import Enum, Flag as _Flag, auto, unique
try:
    from enum import EnumType
except ImportError: # Python<3.11
    from enum import EnumMeta as EnumType

class _StrEnumMeta(EnumType):
    """Python<3.12 requires this in a metaclass (rather than directly in StrEnum)"""
    def __contains__(self, item: str) -> bool:
        try:
            key = item.upper()
        except (AttributeError, TypeError):
            return False
        return key in self.__members__ or item in self.__members__.values()


@unique
class StrEnum(str, Enum, metaclass=_StrEnumMeta):
    """Case-insensitive StrEnum"""
    @classmethod
    def _missing_(cls, value: str):
        return cls.__members__.get(value.upper(), None)

    def __eq__(self, value: str) -> bool:
        try:
            value = self.__class__[value.upper()]
        except (KeyError, ValueError, AttributeError):
            pass
        return super().__eq__(value)

    def __hash__(self) -> int:
        """consistent hashing for dictionary keys"""
        return hash(self.value)

    # compatibility with Python>=3.11 `enum.StrEnum`
    __str__ = str.__str__
    __format__ = str.__format__

    @staticmethod
    def _generate_next_value_(name: str, start, count, last_values) -> str:
        return name.lower()


class Backend(StrEnum):
    """
    Available backends for CIL.

    Examples
    --------
    ```
    FBP(data, backend=Backend.ASTRA)
    FBP(data, backend="astra")
    ```
    """
    ASTRA = auto()
    TIGRE = auto()
    CIL = auto()


class _DimensionBase:
    @classmethod
    def _default_order(cls, engine: str) -> tuple:
        raise NotImplementedError

    @classmethod
    def get_order_for_engine(cls, engine: str, geometry=None) -> tuple:
        """
        Returns the order of dimensions for a specific engine and geometry.

        Parameters
        ----------
        geometry: ImageGeometry | AcquisitionGeometry
            If unspecified, the default order is returned.
        
        Notes
        -----
        In the case that geometry is None, the default order assumes the geometry is not CONE_FLEX, so includes the label 'angle'.
        If you have CONE_FLEX geometry and geometry=None, the returned order then needs to be manually replaced with 'projection',
        or instead pass the geometry to this method to get the correct labels.
        """
        order = cls._default_order(engine)
        if geometry is None:
            return order

        try:
            if geometry.geom_type & AcquisitionType.CONE_FLEX:
                order = [label if label != AcquisitionDimension.ANGLE else AcquisitionDimension.PROJECTION for label in order]
        except AttributeError:
            # if geometry is an ImageGeometry, it does not have a geom_type attribute and we also don't need to adjust the labels
            pass
        return tuple(label for label in order if label in geometry.dimension_labels)

    @classmethod
    def check_order_for_engine(cls, engine: str, geometry) -> bool:
        """
        Returns True iff the order of dimensions is correct for a specific engine and geometry.

        Parameters
        ----------
        geometry: ImageGeometry | AcquisitionGeometry

        Raises
        ------
        ValueError if the order of dimensions is incorrect.
        """
        order_requested = cls.get_order_for_engine(engine, geometry)
        if order_requested == tuple(geometry.dimension_labels):
            return True
        raise ValueError(
            f"Expected dimension_label order {order_requested},"
            f" got {tuple(geometry.dimension_labels)}."
            f" Try using `data.reorder('{engine}')` to permute for {engine}")


class ImageDimension(_DimensionBase, StrEnum):
    """
    Available dimension labels for image data.

    Examples
    --------
    >>> data.reorder([ImageDimension.HORIZONTAL_X, ImageDimension.VERTICAL])
    >>> data.reorder(["horizontal_x", "vertical"])

    """
    CHANNEL = auto()
    VERTICAL = auto()
    HORIZONTAL_X = auto()
    HORIZONTAL_Y = auto()

    @classmethod
    def _default_order(cls, engine: str) -> tuple:
        engine = Backend(engine)
        orders = {
            Backend.ASTRA: (cls.CHANNEL, cls.VERTICAL, cls.HORIZONTAL_Y, cls.HORIZONTAL_X),
            Backend.TIGRE: (cls.CHANNEL, cls.VERTICAL, cls.HORIZONTAL_Y, cls.HORIZONTAL_X),
            Backend.CIL: (cls.CHANNEL, cls.VERTICAL, cls.HORIZONTAL_Y, cls.HORIZONTAL_X)}
        return orders[engine]


class AcquisitionDimension(_DimensionBase, StrEnum):
    """
    Available dimension labels for acquisition data.

    Examples
    --------
    >>> data.reorder([AcquisitionDimension.CHANNEL,
                  AcquisitionDimension.ANGLE,
                  AcquisitionDimension.HORIZONTAL])
    >>> data.reorder(["channel", "angle", "horizontal"])
    """
    CHANNEL = auto()
    ANGLE = auto()
    VERTICAL = auto()
    HORIZONTAL = auto()
    PROJECTION = auto()

    @classmethod
    def _default_order(cls, engine: str) -> tuple:
        engine = Backend(engine)
        orders = {
            Backend.ASTRA: (cls.CHANNEL, cls.VERTICAL, cls.ANGLE, cls.HORIZONTAL),
            Backend.TIGRE: (cls.CHANNEL, cls.ANGLE, cls.VERTICAL, cls.HORIZONTAL),
            Backend.CIL: (cls.CHANNEL, cls.ANGLE, cls.VERTICAL, cls.HORIZONTAL)}
        return orders[engine]


class FillType(StrEnum):
    """
    Available fill types for image data.

    Attributes
    ----------
    RANDOM:
        Fill with random values.
    RANDOM_INT:
        Fill with random integers.
    RANDOM_DEPRECATED:
        Fill with random values using numpy.random.random_sample method
    RANDOM_INT_DEPRECATED:
        Fill with random integers using numpy.random.randint method

    Examples
    --------
    >>> data.fill(FillType.RANDOM)
    >>> data.fill("random")
    """
    RANDOM = auto()
    RANDOM_INT = auto()
    RANDOM_DEPRECATED = auto()
    RANDOM_INT_DEPRECATED = auto()


class AngleUnit(StrEnum):
    """
    Available units for angles.

    Examples
    --------
    >>> data.geometry.set_angles(angle_data, angle_units=AngleUnit.DEGREE)
    >>> data.geometry.set_angles(angle_data, angle_units="degree")

    """
    DEGREE = auto()
    RADIAN = auto()


class _FlagMeta(EnumType):
    """Python<3.12 requires this in a metaclass (rather than directly in Flag)"""
    def __contains__(self, item) -> bool:
        return item.upper() in self.__members__ if isinstance(item, str) else super().__contains__(item)


@unique
class Flag(_Flag, metaclass=_FlagMeta):
    """Case-insensitive Flag"""
    @classmethod
    def _missing_(cls, value):
        return cls.__members__.get(value.upper(), None) if isinstance(value, str) else super()._missing_(value)

    def __eq__(self, value: str) -> bool:
        return super().__eq__(self.__class__(value.upper()) if isinstance(value, str) else value)


class AcquisitionType(Flag):
    """
    Available acquisition types & dimensions.

    WARNING: It's best to use strings rather than integers to initialise.
    >>> AcquisitionType(3) == AcquisitionType(2 | 1) == AcquisitionType.CONE|PARALLEL != AcquisitionType('3D')

    Attributes
    ----------
    PARALLEL:
        Parallel beam.
    CONE:
        Cone beam.
    DIM2:
        2D acquisition.
    DIM3:
        3D acquisition.
    """
    PARALLEL = auto()
    CONE = auto()
    DIM2 = auto()
    DIM3 = auto()
    CONE_FLEX = auto()

    def validate(self):
        """
        Check if the geometry and dimension types are allowed
        """
        assert len(self.dimension) < 2, f"{self} must be 2D xor 3D"
        assert len(self.geometry) < 2, f"{self} must be parallel xor cone beam"
        return self

    @property
    def dimension(self):
        """
        Returns the label for the dimension type
        """
        return self & (self.DIM2 | self.DIM3)

    @property
    def geometry(self):
        """
        Returns the label for the geometry type
        """
        return self & (self.PARALLEL | self.CONE | self.CONE_FLEX)

    @classmethod
    def _missing_(cls, value):
        """2D/3D aliases"""
        if isinstance(value, str):
            value = {'2D': 'DIM2', '3D': 'DIM3'}.get(value.upper(), value)
        return super()._missing_(value)

    def __str__(self) -> str:
        """2D/3D special handling"""
        return '2D' if self == self.DIM2 else '3D' if self == self.DIM3 else (self.name or super().__str__())

    def __hash__(self) -> int:
        """consistent hashing for dictionary keys"""
        return hash(self.value)

    # compatibility with Python>=3.11 `enum.Flag`
    def __len__(self) -> int:
        return bin(self.value).count('1')



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import functools
import warnings
from numbers import Number

import numpy

from ..utilities.multiprocessing import NUM_THREADS
from .labels import FillType


class BackwardCompat(type):
    @property
    def RANDOM(cls):
        warnings.warn("use FillType.RANDOM instead", DeprecationWarning, stacklevel=2)
        return FillType.RANDOM

    @property
    def RANDOM_INT(cls):
        warnings.warn("use FillType.RANDOM_INT instead", DeprecationWarning, stacklevel=2)
        return FillType.RANDOM_INT


class BlockGeometry(metaclass=BackwardCompat):
    @property
    def dtype(self):
        return tuple(i.dtype for i in self.geometries)

    '''Class to hold Geometry as column vector'''
    #__array_priority__ = 1
    def __init__(self, *args, **kwargs):
        ''''''
        self.geometries = args
        self.index = 0
        shape = (len(args),1)
        self.shape = shape

        n_elements = functools.reduce(lambda x,y: x*y, shape, 1)
        if len(args) != n_elements:
            raise ValueError(
                    'Dimension and size do not match: expected {} got {}'
                    .format(n_elements, len(args)))

    def get_item(self, index):
        '''returns the Geometry in the BlockGeometry located at position index'''
        return self.geometries[index]

    def allocate(self, value=0, **kwargs):

        '''Allocates a BlockDataContainer according to geometries contained in the BlockGeometry'''

        symmetry = kwargs.get('symmetry',False)
        containers = [geom.allocate(value, **kwargs) for geom in self.geometries]

        if symmetry == True:

            # for 2x2
            # [ ig11, ig12\
            #   ig21, ig22]

            # Row-wise Order

            if len(containers)==4:
                containers[1]=containers[2]

            # for 3x3
            # [ ig11, ig12, ig13\
            #   ig21, ig22, ig23\
            #   ig31, ig32, ig33]

            elif len(containers)==9:
                containers[1]=containers[3]
                containers[2]=containers[6]
                containers[5]=containers[7]

            # for 4x4
            # [ ig11, ig12, ig13, ig14\
            #   ig21, ig22, ig23, ig24\ c
            #   ig31, ig32, ig33, ig34
            #   ig41, ig42, ig43, ig44]

            elif len(containers) == 16:
                containers[1]=containers[4]
                containers[2]=containers[8]
                containers[3]=containers[12]
                containers[6]=containers[9]
                containers[7]=containers[10]
                containers[11]=containers[15]

        return BlockDataContainer(*containers)

    def __iter__(self):
        '''BlockGeometry is an iterable'''
        return self

    def __next__(self):
        '''BlockGeometry is an iterable'''
        if self.index < len(self.geometries):
            result = self.geometries[self.index]
            self.index += 1
            return result
        else:
            self.index = 0
            raise StopIteration

    def __eq__(self, value: object) -> bool:
        if len(self.geometries) != len(value.geometries):
            return False
        return functools.reduce(lambda x,y: x and y, \
                                [sel == vel for sel,vel in zip(self.geometries, value.geometries)], True)

class BlockDataContainer(object):
    '''Class to hold DataContainers as column vector

    Provides basic algebra between BlockDataContainer's, DataContainer's and
    subclasses and Numbers

    1) algebra between `BlockDataContainer`s will be element-wise, only if
       the shape of the 2 `BlockDataContainer`s is the same, otherwise it
       will fail
    2) algebra between `BlockDataContainer`s and `list` or `numpy array` will
       work as long as the number of `rows` and element of the arrays match,
       independently on the fact that the `BlockDataContainer` could be nested
    3) algebra between `BlockDataContainer` and one `DataContainer` is possible.
       It will require all the `DataContainers` in the block to be
       compatible with the `DataContainer` we want to operate with.
    4) algebra between `BlockDataContainer` and a `Number` is possible and it
       will be done with each element of the `BlockDataContainer` even if nested

    A = [ [B,C] , D]
    A * 3 = [ 3 * [B,C] , 3* D] = [ [ 3*B, 3*C]  , 3*D ]

    '''
    ADD       = 'add'
    SUBTRACT  = 'subtract'
    MULTIPLY  = 'multiply'
    DIVIDE    = 'divide'
    POWER     = 'power'
    SAPYB     = 'sapyb'
    MAXIMUM   = 'maximum'
    MINIMUM   = 'minimum'
    ABS       = 'abs'
    SIGN      = 'sign'
    SQRT      = 'sqrt'
    CONJUGATE = 'conjugate'
    __array_priority__ = 1
    __container_priority__ = 2

    @property
    def dtype(self):
        return tuple(i.dtype for i in self.containers)

    def __init__(self, *args, **kwargs):
        ''''''
        self.containers = args
        self.index = 0
        #if len(set([i.shape for i in self.containers])):
        #    self.geometry = self.containers[0].geometry

        shape = kwargs.get('shape', None)
        if shape is None:
           shape = (len(args),1)
#        shape = (len(args),1)
        self.shape = shape

        n_elements = functools.reduce(lambda x,y: x*y, shape, 1)
        if len(args) != n_elements:
            raise ValueError(
                    'Dimension and size do not match: expected {} got {}'
                    .format(n_elements, len(args)))


    def __iter__(self):
        '''BlockDataContainer is Iterable'''
        self.index=0
        return self
    def next(self):
        '''python2 backwards compatibility'''
        return self.__next__()
    def __next__(self):
        try:
            out = self[self.index]
        except IndexError as ie:
            raise StopIteration()
        self.index+=1
        return out

    def is_compatible(self, other):
        '''basic check if the size of the 2 objects fit'''

        if isinstance(other, Number):
            return True
        elif isinstance(other, (list, tuple, numpy.ndarray)) :
            for ot in other:
                if not isinstance(ot, Number):
                    raise ValueError('List/ numpy array can only contain numbers {}'\
                                     .format(type(ot)))
            return len(self.containers) == len(other)
        elif isinstance(other, BlockDataContainer):
            return len(self.containers) == len(other.containers)
        else:
            # this should work for other as DataContainers and children
            ret = True
            for i, el in enumerate(self.containers):
                if isinstance(el, BlockDataContainer):
                    a = el.is_compatible(other)
                else:
                    a = el.shape == other.shape
                ret = ret and a
            # probably will raise
            return ret


    def get_item(self, row):
        if row > self.shape[0]:
            raise ValueError('Requested row {} > max {}'.format(row, self.shape[0]))
        return self.containers[row]

    def __getitem__(self, row):
        return self.get_item(row)

    def add(self, other, *args, **kwargs):
        '''Algebra: add method of BlockDataContainer with number/DataContainer or BlockDataContainer

        Parameters
        ----------
        other : number, DataContainer or subclasses or BlockDataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result
        '''
        return self.binary_operations(BlockDataContainer.ADD, other, *args, **kwargs)
    def subtract(self, other, *args, **kwargs):
        '''Algebra: subtract method of BlockDataContainer with number/DataContainer or BlockDataContainer

        Parameters
        ----------
        other : number, DataContainer or subclasses or BlockDataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result
        '''
        return self.binary_operations(BlockDataContainer.SUBTRACT, other, *args, **kwargs)
    def multiply(self, other, *args, **kwargs):
        '''Algebra: multiply method of BlockDataContainer with number/DataContainer or BlockDataContainer

        Parameters
        ----------
        other : number, DataContainer or subclasses or BlockDataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result
        '''
        return self.binary_operations(BlockDataContainer.MULTIPLY, other, *args, **kwargs)
    def divide(self, other, *args, **kwargs):
        '''Algebra: divide method of BlockDataContainer with number/DataContainer or BlockDataContainer

        Parameters
        ----------
        other : number, DataContainer or subclasses or BlockDataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result

        '''
        return self.binary_operations(BlockDataContainer.DIVIDE, other, *args, **kwargs)
    def power(self, other, *args, **kwargs):
        '''Algebra: power method of BlockDataContainer with number/DataContainer or BlockDataContainer

        Parameters
        ----------
        other : number, DataContainer or subclasses or BlockDataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result
        '''
        return self.binary_operations(BlockDataContainer.POWER, other, *args, **kwargs)
    def maximum(self, other, *args, **kwargs):
        '''Algebra: maximum method of BlockDataContainer with number/DataContainer or BlockDataContainer

        Parameters
        ----------
        other : number, DataContainer or subclasses or BlockDataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result
        '''
        return self.binary_operations(BlockDataContainer.MAXIMUM, other, *args, **kwargs)
    def minimum(self, other, *args, **kwargs):
        '''Algebra: minimum method of BlockDataContainer with number/DataContainer or BlockDataContainer

        Parameters
        ----------
        other : number, DataContainer or subclasses or BlockDataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result

        '''
        return self.binary_operations(BlockDataContainer.MINIMUM, other, *args, **kwargs)

    def sapyb(self, a, y, b, out=None, num_threads = NUM_THREADS):
        r'''performs axpby element-wise on the BlockDataContainer containers

        Does the operation .. math:: a*x+b*y and stores the result in out, where x is self

        Parameters
        ----------
        a : scalar or BlockDataContainer
        b : scalar or BlockDataContainer
        y : compatible (Block)DataContainer
        out : BlockDataContainer, optional
            Provides a placeholder for the result

        Example
        -------
        >>> a = 2
        >>> b = 3
        >>> ig = ImageGeometry(10,11)
        >>> x = ig.allocate(1)
        >>> y = ig.allocate(2)
        >>> bdc1 = BlockDataContainer(2*x, y)
        >>> bdc2 = BlockDataContainer(x, 2*y)
        >>> out = bdc1.sapyb(a,bdc2,b)
        '''
        if out is None:
            out = self * 0
        kwargs = {'a':a, 'b':b, 'out':out, 'num_threads': NUM_THREADS}
        return self.binary_operations(BlockDataContainer.SAPYB, y, **kwargs)


    def binary_operations(self, operation, other, *args, **kwargs):
        '''Algebra: generic method of algebric operation with BlockDataContainer with number/DataContainer or BlockDataContainer

        Provides commutativity with DataContainer and subclasses, i.e. this
        class's reverse algebraic methods take precedence w.r.t. direct algebraic
        methods of DataContainer and subclasses.

        This method is not to be used directly
        '''
        if not self.is_compatible(other):
            raise ValueError('Incompatible for operation {}'.format(operation))
        out = kwargs.get('out', None)
        if isinstance(other, Number):
            # try to do algebra with one DataContainer. Will raise error if not compatible
            kw = kwargs.copy()
            res = []
            for i,el in enumerate(self.containers):
                if operation == BlockDataContainer.ADD:
                    op = el.add
                elif operation == BlockDataContainer.SUBTRACT:
                    op = el.subtract
                elif operation == BlockDataContainer.MULTIPLY:
                    op = el.multiply
                elif operation == BlockDataContainer.DIVIDE:
                    op = el.divide
                elif operation == BlockDataContainer.POWER:
                    op = el.power
                elif operation == BlockDataContainer.MAXIMUM:
                    op = el.maximum
                elif operation == BlockDataContainer.MINIMUM:
                    op = el.minimum
                else:
                    raise ValueError('Unsupported operation', operation)
                if out is not None:
                    kw['out'] = out.get_item(i)
                    op(other, *args, **kw)
                else:
                    res.append(op(other, *args, **kw))
            if out is not None:
                return out
            else:
                return type(self)(*res, shape=self.shape)
        elif isinstance(other, (list, tuple, numpy.ndarray, BlockDataContainer)):
            kw = kwargs.copy()
            res = []
            if isinstance(other, BlockDataContainer):
                the_other = other.containers
            else:
                the_other = other

            for i,zel in enumerate(zip ( self.containers, the_other) ):
                el = zel[0]
                ot = zel[1]
                if operation == BlockDataContainer.ADD:
                    op = el.add
                elif operation == BlockDataContainer.SUBTRACT:
                    op = el.subtract
                elif operation == BlockDataContainer.MULTIPLY:
                    op = el.multiply
                elif operation == BlockDataContainer.DIVIDE:
                    op = el.divide
                elif operation == BlockDataContainer.POWER:
                    op = el.power
                elif operation == BlockDataContainer.MAXIMUM:
                    op = el.maximum
                elif operation == BlockDataContainer.MINIMUM:
                    op = el.minimum
                elif operation == BlockDataContainer.SAPYB:
                    if not isinstance(other, BlockDataContainer):
                        raise ValueError("{} cannot handle {}".format(operation, type(other)))
                    op = el.sapyb
                else:
                    raise ValueError('Unsupported operation', operation)

                if out is not None:
                    if operation == BlockDataContainer.SAPYB:
                        if isinstance(kw['a'], BlockDataContainer):
                            a = kw['a'].get_item(i)
                        else:
                            a = kw['a']

                        if isinstance(kw['b'], BlockDataContainer):
                            b = kw['b'].get_item(i)
                        else:
                            b = kw['b']

                        el.sapyb(a, ot, b, out.get_item(i), num_threads=kw['num_threads'])
                    else:
                        kw['out'] = out.get_item(i)
                        op(ot, *args, **kw)
                else:
                    res.append(op(ot, *args, **kw))
            if out is not None:
                return out
            else:
                return type(self)(*res, shape=self.shape)
        else:
            # try to do algebra with one DataContainer. Will raise error if not compatible
            kw = kwargs.copy()
            if operation != BlockDataContainer.SAPYB:
                # remove keyworded argument related to SAPYB
                for k in ['a','b','y', 'num_threads', 'dtype']:
                    if k in kw.keys():
                        kw.pop(k)

            res = []
            for i,el in enumerate(self.containers):
                if operation == BlockDataContainer.ADD:
                    op = el.add
                elif operation == BlockDataContainer.SUBTRACT:
                    op = el.subtract
                elif operation == BlockDataContainer.MULTIPLY:
                    op = el.multiply
                elif operation == BlockDataContainer.DIVIDE:
                    op = el.divide
                elif operation == BlockDataContainer.POWER:
                    op = el.power
                elif operation == BlockDataContainer.MAXIMUM:
                    op = el.maximum
                elif operation == BlockDataContainer.MINIMUM:
                    op = el.minimum
                elif operation == BlockDataContainer.SAPYB:

                    if isinstance(kw['a'], BlockDataContainer):
                        a = kw['a'].get_item(i)
                    else:
                        a = kw['a']

                    if isinstance(kw['b'], BlockDataContainer):
                        b = kw['b'].get_item(i)
                    else:
                        b = kw['b']

                    el.sapyb(a, other, b, out.get_item(i), kw['num_threads'])

                    # As axpyb cannot return anything we `continue` to skip the rest of the code block
                    continue

                else:
                    raise ValueError('Unsupported operation', operation)
                if out is not None:
                    kw['out'] = out.get_item(i)
                    op(other, *args, **kw)
                else:
                    res.append(op(other, *args, **kw))

            if out is not None:
                return out
            else:
                return type(self)(*res, shape=self.shape)

    ## unary operations

    def unary_operations(self, operation, *args, **kwargs ):
        '''Unary operation on BlockDataContainer:

        generic method of unary operation with BlockDataContainer: abs, sign, sqrt and conjugate

        This method is not to be used directly
        '''
        out = kwargs.get('out', None)
        kw = kwargs.copy()
        if out is None:
            res = []
            for el in self.containers:
                if operation == BlockDataContainer.ABS:
                    op = el.abs
                elif operation == BlockDataContainer.SIGN:
                    op = el.sign
                elif operation == BlockDataContainer.SQRT:
                    op = el.sqrt
                elif operation == BlockDataContainer.CONJUGATE:
                    op = el.conjugate
                res.append(op(*args, **kw))
            return BlockDataContainer(*res)
        else:
            kw.pop('out')
            for el,elout in zip(self.containers, out.containers):
                if operation == BlockDataContainer.ABS:
                    op = el.abs
                elif operation == BlockDataContainer.SIGN:
                    op = el.sign
                elif operation == BlockDataContainer.SQRT:
                    op = el.sqrt
                elif operation == BlockDataContainer.CONJUGATE:
                    op = el.conjugate
                kw['out'] = elout
                op(*args, **kw)

    def abs(self, *args, **kwargs):
        return self.unary_operations(BlockDataContainer.ABS, *args, **kwargs)
    def sign(self, *args, **kwargs):
        return self.unary_operations(BlockDataContainer.SIGN, *args, **kwargs)
    def sqrt(self, *args, **kwargs):
        return self.unary_operations(BlockDataContainer.SQRT, *args, **kwargs)
    def conjugate(self, *args, **kwargs):
        return self.unary_operations(BlockDataContainer.CONJUGATE, *args, **kwargs)
    # def abs(self, *args,  **kwargs):
    #     return type(self)(*[ el.abs(*args, **kwargs) for el in self.containers], shape=self.shape)
    # def sign(self, *args,  **kwargs):
    #     return type(self)(*[ el.sign(*args, **kwargs) for el in self.containers], shape=self.shape)
    # def sqrt(self, *args,  **kwargs):
    #     return type(self)(*[ el.sqrt(*args, **kwargs) for el in self.containers], shape=self.shape)
    # def conjugate(self, out=None):
    #     return type(self)(*[el.conjugate() for el in self.containers], shape=self.shape)

    ## reductions

    def sum(self, *args, **kwargs):
        return numpy.sum([ el.sum(*args, **kwargs) for el in self.containers])

    def squared_norm(self):
        y = numpy.asarray([el.squared_norm() for el in self.containers])
        return y.sum()


    def norm(self):
        return numpy.sqrt(self.squared_norm())

    def pnorm(self, p=2):
        # See https://github.com/TomographicImaging/CIL/issues/1525#issuecomment-1757413803
        if not functools.reduce(lambda x,y: x and y, [el.shape == self.containers[0].shape for el in self.containers], True):
            raise ValueError('pnorm: Incompatible shapes - each container in the BlockDataContainer must have the same shape in order to calculate the pnorm')
        if p==1:
            return sum(self.abs())
        elif p==2:
            tmp = functools.reduce(lambda a,b: a + b.conjugate()*b, self.containers, self.get_item(0) * 0 ).sqrt()
            return tmp
        else:
            return ValueError('Not implemented')

    def copy(self):
        '''alias of clone'''
        return self.clone()
    def clone(self):
        return type(self)(*[el.copy() for el in self.containers], shape=self.shape)
    def fill(self, other):
        if isinstance (other, BlockDataContainer):
            if not self.is_compatible(other):
                raise ValueError('Incompatible containers')
            for el,ot in zip(self.containers, other.containers):
                el.fill(ot)
        else:
            return ValueError('Cannot fill with object provided {}'.format(type(other)))

    def __add__(self, other):
        return self.add( other )
    # __radd__

    def __sub__(self, other):
        return self.subtract( other )
    # __rsub__

    def __mul__(self, other):
        return self.multiply(other)
    # __rmul__

    def __div__(self, other):
        return self.divide(other)
    # __rdiv__
    def __truediv__(self, other):
        return self.divide(other)

    def __pow__(self, other):
        return self.power(other)
    # reverse operand
    def __radd__(self, other):
        '''Reverse addition

        to make sure that this method is called rather than the __mul__ of a numpy array
        the class constant __array_priority__ must be set > 0
        https://docs.scipy.org/doc/numpy-1.15.1/reference/arrays.classes.html#numpy.class.__array_priority__
        '''
        return self + other
    # __radd__

    def __rsub__(self, other):
        '''Reverse subtraction

        to make sure that this method is called rather than the __mul__ of a numpy array
        the class constant __array_priority__ must be set > 0
        https://docs.scipy.org/doc/numpy-1.15.1/reference/arrays.classes.html#numpy.class.__array_priority__
        '''
        return (-1 * self) + other
    # __rsub__

    def __rmul__(self, other):
        '''Reverse multiplication

        to make sure that this method is called rather than the __mul__ of a numpy array
        the class constant __array_priority__ must be set > 0
        https://docs.scipy.org/doc/numpy-1.15.1/reference/arrays.classes.html#numpy.class.__array_priority__
        '''
        return self * other
    # __rmul__

    def __rdiv__(self, other):
        '''Reverse division

        to make sure that this method is called rather than the __mul__ of a numpy array
        the class constant __array_priority__ must be set > 0
        https://docs.scipy.org/doc/numpy-1.15.1/reference/arrays.classes.html#numpy.class.__array_priority__
        '''
        return pow(self / other, -1)
    # __rdiv__
    def __rtruediv__(self, other):
        '''Reverse truedivision

        to make sure that this method is called rather than the __mul__ of a numpy array
        the class constant __array_priority__ must be set > 0
        https://docs.scipy.org/doc/numpy-1.15.1/reference/arrays.classes.html#numpy.class.__array_priority__
        '''
        return self.__rdiv__(other)

    def __rpow__(self, other):
        '''Reverse power

        to make sure that this method is called rather than the __mul__ of a numpy array
        the class constant __array_priority__ must be set > 0
        https://docs.scipy.org/doc/numpy-1.15.1/reference/arrays.classes.html#numpy.class.__array_priority__
        '''
        return other.power(self)

    def __iadd__(self, other):
        '''Inline addition'''
        if isinstance (other, BlockDataContainer):
            for el,ot in zip(self.containers, other.containers):
                el += ot
        elif isinstance(other, Number):
            for el in self.containers:
                el += other
        elif isinstance(other, list) or isinstance(other, numpy.ndarray):
            if not self.is_compatible(other):
                raise ValueError('Incompatible for __iadd__')
            for el,ot in zip(self.containers, other):
                el += ot
        return self
    # __iadd__

    def __isub__(self, other):
        '''Inline subtraction'''
        if isinstance (other, BlockDataContainer):
            for el,ot in zip(self.containers, other.containers):
                el -= ot
        elif isinstance(other, Number):
            for el in self.containers:
                el -= other
        elif isinstance(other, list) or isinstance(other, numpy.ndarray):
            if not self.is_compatible(other):
                raise ValueError('Incompatible for __isub__')
            for el,ot in zip(self.containers, other):
                el -= ot
        return self
    # __isub__

    def __imul__(self, other):
        '''Inline multiplication'''
        if isinstance (other, BlockDataContainer):
            for el,ot in zip(self.containers, other.containers):
                el *= ot
        elif isinstance(other, Number):
            for el in self.containers:
                el *= other
        elif isinstance(other, list) or isinstance(other, numpy.ndarray):
            if not self.is_compatible(other):
                raise ValueError('Incompatible for __imul__')
            for el,ot in zip(self.containers, other):
                el *= ot
        return self
    # __imul__

    def __idiv__(self, other):
        '''Inline division'''
        if isinstance (other, BlockDataContainer):
            for el,ot in zip(self.containers, other.containers):
                el /= ot
        elif isinstance(other, Number):
            for el in self.containers:
                el /= other
        elif isinstance(other, list) or isinstance(other, numpy.ndarray):
            if not self.is_compatible(other):
                raise ValueError('Incompatible for __idiv__')
            for el,ot in zip(self.containers, other):
                el /= ot
        return self
    # __rdiv__
    def __itruediv__(self, other):
        '''Inline truedivision'''
        return self.__idiv__(other)

    def __neg__(self):
        """ Return - self """
        return -1 * self

    def dot(self, other):
#
        tmp = [ self.containers[i].dot(other.containers[i]) for i in range(self.shape[0])]
        return sum(tmp)

    def __len__(self):

        return self.shape[0]

    @property
    def geometry(self):
        try:
            return BlockGeometry(*[el.geometry.copy() for el in self.containers])
        except AttributeError:
            return None



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# Kyle Pidgeon (UKRI-STFC)
# Hussam Alhassan (UKRI-STFC)


import matplotlib.lines as mlines
import matplotlib.pyplot as plt
from matplotlib.patches import FancyArrowPatch

from cil.framework import AcquisitionGeometry, AcquisitionData, ImageData, DataContainer, BlockDataContainer
from cil.framework.labels import AcquisitionType
import numpy as np
import warnings

import os
from mpl_toolkits.mplot3d import proj3d
from mpl_toolkits.axes_grid1 import make_axes_locatable
from itertools import cycle
import logging

log = logging.getLogger(__name__)


class _PlotData(object):
    def __init__(self, data, title, axis_labels, origin):
        self.data = data
        self.title = title
        self.axis_labels = axis_labels
        self.origin = origin
        self.range = None

def set_origin(data, origin):
    shape_v = [0, data.shape[0]]
    shape_h = [0, data.shape[1]]

    if type(data) != np.ndarray:
        data = data.as_array()

    data_origin='lower'

    if 'upper' in origin:
        shape_v.reverse()
        data_origin='upper'

    if 'right' in origin:
        shape_h.reverse()
        data = np.flip(data,1)

    extent = (*shape_h,*shape_v)
    return data, data_origin, extent

class show_base(object):
    def save(self,filename, **kwargs):
        '''
        Saves the image using matplotlib.figure.savefig(). Default format is png if no extension is provided.

        matplotlib kwargs can be passed, refer to documentation
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html
        '''

        file,extension = os.path.splitext(os.path.abspath(filename))
        extension = extension.strip('.')

        extensions = plt.gcf().canvas.get_supported_filetypes()
        extensions = [i for i in extensions.keys()]

        format = kwargs.get('format',None)

        if format is None:
            if extension == '':
                extension = 'png'
        else:
            extension = format

        if extension not in extensions:
            raise ValueError("Extension not valid. Got {0}, backend supports {1}".format(extension,extensions))

        try:
            path_full = file+'.'+extension
            self.figure.set_tight_layout(True)
            self.figure.set_facecolor('w')
            self.figure.savefig(path_full, bbox_inches='tight',**kwargs)
            print("Saved image as {}".format(path_full))
        except PermissionError:
            print("Unable to save image. Permissions denied: {}".format(path_full))
        except:
            print("Unable to save image")


class show1D(show_base):
    """
    Displays the 1D line profile of a N-dimensional dataset along a specified axis.

    If multiple slices are provided, a plot will be generated for each slice.

    If multiple datasets are provided, the same slice will be applied to each dataset.

    Parameters
    ----------
    data : ndarray i.e. DataContainer, numpy.ndarray or a list of ndarrays
        The data to be sliced and plotted. If a list of DataContainers is passed, they must all have the same shape.
    slice_list : list of tuples, or list of list of tuple, optional.
        Describes the slice to be taken from the data to extract a 1D vector. If None, the centre line of each axis will be displayed.
        Each slice should be a list of tuples, where each tuple is a (axis, index). Each axis can be described as a string or integer representing the axis to slice along.
        To plot multiple slices, pass a list of slices.        
    dataset_labels : list of str, optional
        Overrides the default legend labels for each dataset displayed in the plot legend. Must be a list of strings equal to the number of datasets.
    title: list of str, optional
        Overrides the default title for each plot. Must be a list of strings equal to the number of plots.
    line_colours : str, list of str, default=None
        Overrides the default colour(s) for each line plot. Must be a list of colours equal to the number of datasets.
    line_styles : {"-","--","-.",":"}, list of {"-","--","-.",":"}, default=None
        Overrides the default line style(s) for each line plot. Must be a list of styles equal to the number of datasets.
    axis_labels : tuple of str, list of tuples, optional
        Overrides the default axis labels in the form (x_axis_label, y_axis_label). Must be a single tuple or a list of tuples equal to the number of plots.
    size : tuple, default=(8,3)
        The size of each sub-plot in the figure.

    Note
    ----

    The figure can be saved using the `save` method . i.e:

    >>> fig = show1D(data, title='My Plot')
    >>> fig.save('/path/to/output.png')
    

    Examples
    --------

    This example uses a 3D dataset with no slicing information provided. The function
    displays 3 subplots with the centre line profile shown along each axis.

    >>> from cil.utilities.display import show1D
    >>> from cil.utilities import dataexample
    >>> data = dataexample.SIMULATED_SPHERE_VOLUME.get()
    >>> show1D(data, slice_list=None)


    The following example shows a single line profile from a 3D dataset along the vertical axis.

    >>> from cil.utilities.display import show1D
    >>> from cil.utilities import dataexample
    >>> data = dataexample.SIMULATED_SPHERE_VOLUME.get()
    >>> show1D(data, slice_list=[("Horizontal_x", 30),("Horizontal_y", 40)])


    The following example shows multiple line profiles from a 3D dataset along the vertical axis.

    >>> from cil.utilities.display import show1D
    >>> from cil.utilities import dataexample
    >>> data = dataexample.SIMULATED_SPHERE_VOLUME.get()
    >>> show1D(data, slice_list=[[("Vertical",62),("Horizontal_y",70)],[("Vertical",61),("Horizontal_y",70)],[("Vertical",67),("Horizontal_x",70)]])


    The following example plots the same line profile from two datasets on the same plot.

    >>> from cil.utilities.display import show1D
    >>> from cil.utilities import dataexample
    >>> data = dataexample.SIMULATED_SPHERE_VOLUME.get()
    >>> data2 = data*0.9
    >>> show1D([data, data2], slice_list=[("Horizontal_x", 30),("Horizontal_y", 40)])

    
    The following example shows the case where the input data is a numpy array.
    The slice_list can be provided as (axis, index), where axis is an integer.
    
    >>> from cil.utilities.display import show1D
    >>> import numpy as np
    >>> data = np.random.rand(10,10,10)
    >>> show1D(data, slice_list=[(0, 3),(2, 5)])


    To save the figure, use the `save` method.

    >>> from cil.utilities.display import show1D
    >>> from cil.utilities import dataexample
    >>> data = dataexample.SIMULATED_SPHERE_VOLUME.get()
    >>> fig = show1D(data, slice_list=None)
    >>> fig.save('/path/to/output.png')

    """

    def __init__(self, data, slice_list=None, dataset_labels=None, title=None,
                 line_colours=None, line_styles=None, axis_labels=None,
                 size=(8,3)):

        self.figure = self._show1d(data, slice_list, dataset_labels=dataset_labels, title=title,
                                   line_colours=line_colours, line_styles=line_styles,
                                   axis_labels=axis_labels, size=size)


    def _parse_slice(self, slice_list_in, items_per_slice, ndim, labels):
        """
        Internal function to parse the slice_list input and convert it to a slice object to be applied to the data.
        """

        # check enough slices are provided to output a single vector
        if len(slice_list_in) != items_per_slice:
            raise ValueError("slice_list must provide a slice for ndim - 1 axes")
        
        # convert axis labels to indices
        axis_indices = []
        for sl in slice_list_in:
            try:
                ind = int(sl[0])
            except:
                if sl[0] in labels:
                    ind = labels.index(sl[0])
                else:
                    raise ValueError(f"Invalid axis label: {sl[0]}")
            axis_indices.append(ind)
            

        # check axis are unique
        if len(set([index for index in axis_indices])) != items_per_slice:
            raise ValueError("slice_list contains duplicate axes. Each axis must be unique.")

        slice_ndim = [slice(None)]*ndim 
        
        for i, sl in enumerate(slice_list_in):
            slice_ndim[axis_indices[i]] = slice(sl[1], sl[1] + 1)
        
        slice_list_out = tuple(slice_ndim)
        
        return slice_list_out

    def _show1d(self, data, slice_list=None, dataset_labels='default', title=None, line_colours=None,
                line_styles=None, axis_labels='default', size=(8,3)):
        """
        Internal function to display 1D plots of pixel flux from multi-dimensional data and slicing information.
        """
       
        # check datasets
        if hasattr(data, 'ndim'):
            data_list = [data]
            num_data = 1
        else:
            data_list = data
            num_data = len(data)

            for i in range(1, num_data):
                if data_list[i].shape != data_list[0].shape:
                    raise ValueError("All datasets must have the same shape")
                
        data_ndim = data_list[0].ndim
        data_shape = data_list[0].shape
        items_per_slice = len(data_shape) - 1

        try:
            labels = data_list[0].dimension_labels
        except:
            labels = [f"Axis {i}" for i in range(data_ndim)]

        if slice_list is None:
            # show the centre slice of each axes
            _slice_list = [None]*data_ndim 

            for i in range(data_ndim):
                slice_ndim = [slice(None)]*data_ndim
                for j in range(data_ndim):
                    if i != j:
                        slice_ndim[j] = slice(data_shape[j]//2, data_shape[j]//2 + 1)

                _slice_list[i] = tuple(slice_ndim)
        
        elif isinstance(slice_list, list):
            if isinstance(slice_list[0], tuple):
                # single slice only
                _slice_list = [None]*1
                _slice_list[0] = self._parse_slice(slice_list, items_per_slice, data_ndim, labels)

            elif isinstance(slice_list[0], list):
                # multiple slices
                _slice_list = [None]*len(slice_list)

                for i in range(len(slice_list)):
                    _slice_list[i]  = self._parse_slice(slice_list[i], items_per_slice, data_ndim, labels)
            else:
                raise ValueError("slice_list must contain a valid list of slices to generate 1D plots")
        
        # set up plots
        CB_PALETTE = ['#377eb8', '#ff7f00', '#4daf4a',
                    '#f781bf', '#a65628', '#984ea3',
                    '#999999', '#e41a1c', '#dede00']
        
        if line_colours is not None and len(line_colours) < num_data:
            log.warning ("line_colours must be a list of colours at least as long as the number of datasets, using default colour palette")
            line_colours = None

        if line_styles is not None and len(line_styles) < num_data:
            log.warning ("line_styles must be a list of styles at least as long as the number of datasets, using default line styles")
            line_styles = None
        
        num_sub_plots = len(_slice_list)
        fig_size = (size[0], size[1] * num_sub_plots)
        fig, (ax) = plt.subplots(num_sub_plots, figsize=fig_size)

        if num_sub_plots == 1:
            ax = np.array(ax)

        for i in range(num_sub_plots):

            #reset the colour and line style
            colour_cyc = cycle(CB_PALETTE)
            ls_cyc = cycle(["-","--","-.",":"])

            # extract the vectors
            for j, x in enumerate(data_list):
                try:
                    x = x.as_array()
                except:
                    pass

                _cl = next(colour_cyc) if line_colours is None else line_colours[j]
                _ls = next(ls_cyc) if line_styles is None else line_styles[j]
             
                data_line = np.squeeze(x[_slice_list[i]])

                if dataset_labels is None:
                    _lbl = f'Dataset {j}'
                elif isinstance(dataset_labels, list) and len(dataset_labels) == num_data:
                    _lbl = dataset_labels[j]
                else:
                    raise ValueError("dataset_labels must be a list of strings equal to the number of datasets")

                ax.flat[i].plot(data_line, color=_cl, linestyle=_ls, label=_lbl)

            # get the unsliced index
            subplot_title = f"Slice at "
            for j, sl in enumerate(_slice_list[i]):
                if sl == slice(None):
                    x_label = f"{labels[j]} index"
                else:
                    subplot_title += f"{labels[j]}:{sl.start}, "

            # remove last two characters
            subplot_title = subplot_title[:-2]

            if title is not None:
                if isinstance(title, list) and len(title) == num_sub_plots:
                    subplot_title = title[i]
                elif isinstance(title, str):
                    subplot_title = title
                else:
                    raise ValueError("title must be a list of strings equal to the number of plots")

            ax.flat[i].set_title(subplot_title)

            if axis_labels is not None:
                if isinstance(axis_labels, tuple):
                    ax.flat[i].set_xlabel(axis_labels[0])
                    ax.flat[i].set_ylabel(axis_labels[1])
                elif isinstance(axis_labels, list) and len(axis_labels) == num_sub_plots:
                    ax.flat[i].set_xlabel(axis_labels[i][0])
                    ax.flat[i].set_ylabel(axis_labels[i][1])
                else:
                    raise ValueError("axis_labels must be a tuple or a list of tuples equal to the number of plots")
            else:
                ax.flat[i].set_xlabel(x_label)
                ax.flat[i].set_ylabel("Value")

            ax.flat[i].set_xlim(0, len(data_line)-1)

            if num_data > 1:
                ax.flat[i].legend(loc='center left', bbox_to_anchor=(1, 0.5))

        fig.set_tight_layout(True)
        fig.set_facecolor('w')

        #plt.show() creates a new figure so we save a copy to return
        fig2 = plt.gcf()
        plt.show()
        return fig2


class show2D(show_base):
    '''This plots 2D slices from cil DataContainer types.

    Plots 1 or more 2D plots in an (n x num_cols) matrix.
    Can plot multiple slices from one 3D dataset, or compare multiple datasets
    Inputs can be single arguments or list of arguments that will be sequentially applied to subplots
    If no slice_list is passed a 3D dataset will display the centre slice of the outer dimension, a 4D dataset will show the centre slices of the two outer dimension.


    Parameters
    ----------
    datacontainers: ImageData, AcquisitionData, list of  ImageData / AcquisitionData, BlockDataContainer
        The DataContainers to be displayed
    title: string, list of strings, optional
        The title for each figure
    slice_list: tuple, int, list of tuples, list of ints, optional
        The slices to show. A list of integers will show slices for the outer dimension. For 3D datacontainers single slice: (direction, index). For 4D datacontainers two slices: [(direction0, index),(direction1, index)].
    fix_range: boolean, tuple, list of tuples
        Sets the display range of the data. `True` sets all plots to the global (min, max).
    axis_labels: tuple, list of tuples, optional
        The axis labels for each figure e.g. ('x','y')
    origin: string, list of strings
        Sets the display origin. 'lower/upper-left/right'
    cmap: str, list or tuple of strings
        Sets the colour map of the plot (see matplotlib.pyplot). If passed a list or tuple of the
        length of datacontainers, allows to set a different color map for each datacontainer.
    num_cols: int
        Sets the number of columns of subplots to display
    size: tuple
        Figure size in inches

        
    Note
    ----

    The figure can be saved using the `save` method . i.e:

    >>> fig = show2D(data, title='My Plot')
    >>> fig.save('/path/to/output.png')

   '''

    def __init__(self,datacontainers, title=None, slice_list=None, fix_range=False, axis_labels=None, origin='lower-left', cmap='gray', num_cols=2, size=(15,15)):

        self.figure = self.__show2D(datacontainers, title=title, slice_list=slice_list, fix_range=fix_range, axis_labels=axis_labels, origin=origin, cmap=cmap, num_cols=num_cols, size=size)

    def __show2D(self,datacontainers, title=None, slice_list=None, fix_range=False, axis_labels=None, origin='lower-left', cmap='gray', num_cols=2, size=(15,15)):

        #get number of subplots, number of input datasets, or number of slices requested
        if isinstance(datacontainers, (list, BlockDataContainer)):
            num_plots = len(datacontainers)
        else:
            dim = len(datacontainers.shape)

            if slice_list is None or dim == 2:
                num_plots = 1
            elif type(slice_list) is tuple:
                num_plots = 1
            elif dim == 4 and type(slice_list[0]) is tuple:
                num_plots = 1
            else:
                num_plots = len(slice_list)

        subplots = []

        #range needs subsetted data
        range_min = float("inf")
        range_max = -range_min

        #set up, all inputs can be 1 or num_plots
        for i in range(num_plots):

            #get data per subplot, subset where required
            if isinstance(datacontainers, (list, BlockDataContainer)):
                data = datacontainers[i]
            else:
                data = datacontainers

            if len(data.shape) ==4:

                if slice_list is None or type(slice_list) is tuple or type(slice_list[0]) is tuple: #none, (direction, ind) or [(direction0, ind), (direction1, ind)] apply to all datasets
                    slice_requested = slice_list
                elif type(slice_list[i]) == int or len(slice_list[i]) > 1: # [ind0, ind1, ind2] of direction0, or [[(direction0, ind), (direction1, ind)],[(direction0, ind), (direction1, ind)]]
                    slice_requested = slice_list[i]
                else:
                    slice_requested = slice_list[i][0] # [[(direction0, ind)],[(direction0, ind)]]

                cut_axis = [0,1]
                cut_slices = [data.shape[0]//2, data.shape[1]//2]

                if type(slice_requested) is int:
                    #use axis 0, slice val
                    cut_slices[0] = slice_requested
                elif type(slice_requested) is tuple:
                    #get axis ind,
                    # if 0 default 1
                    # if 1 default 0
                    axis = slice_requested[0]
                    if slice_requested[0] is str:
                        axis = data.dimension_labels.index(axis)

                    if axis == 0:
                        cut_axis[0] = slice_requested[0]
                        cut_slices[0] = slice_requested[1]
                    else:
                        cut_axis[1] = slice_requested[0]
                        cut_slices[1] = slice_requested[1]

                elif type(slice_requested) is list:
                    #use full input
                    cut_axis[0] = slice_requested[0][0]
                    cut_axis[1] = slice_requested[1][0]
                    cut_slices[0] = slice_requested[0][1]
                    cut_slices[1] = slice_requested[1][1]

                    if cut_axis[0] > cut_axis[1]:
                        cut_axis.reverse()
                        cut_slices.reverse()

                try:
                    if hasattr(data, 'get_slice'):
                        if type(cut_axis[0]) is int:
                            cut_axis[0] = data.dimension_labels[cut_axis[0]]
                        if type(cut_axis[1]) is int:
                            cut_axis[1] = data.dimension_labels[cut_axis[1]]

                        temp_dict = {cut_axis[0]:cut_slices[0], cut_axis[1]:cut_slices[1]}
                        plot_data = data.get_slice(**temp_dict, force=True)
                    elif hasattr(data,'as_array'):
                        plot_data = data.as_array().take(indices=cut_slices[1], axis=cut_axis[1])
                        plot_data = plot_data.take(indices=cut_slices[0], axis=cut_axis[0])
                    else:
                        plot_data = data.take(indices=cut_slices[1], axis=cut_axis[1])
                        plot_data = plot_data.take(indices=cut_slices[0], axis=cut_axis[0])

                except:
                    raise TypeError("Unable to slice input data. Could not obtain 2D slice {0} from {1} with shape {2}.\n\
                            Pass either correct slice information or a 2D array".format(slice_requested, type(data), data.shape))

                subtitle = "direction: ({0},{1}), slice: ({2},{3})".format(*cut_axis, * cut_slices)


            elif len(data.shape) == 3:
                #get slice list per subplot
                if type(slice_list) is list:            #[(direction, ind), (direction, ind)],  [ind0, ind1, ind2] of direction0
                    slice_requested = slice_list[i]
                else:                                   #(direction, ind) single tuple apply to all datasets
                    slice_requested = slice_list

                #default axis 0, centre slice
                cut_slice = data.shape[0]//2
                cut_axis = 0

                if type(slice_requested) is int:
                    #use axis 0, slice val
                    cut_slice = slice_requested
                if type(slice_requested) is tuple:
                    cut_slice = slice_requested[1]
                    cut_axis = slice_requested[0]

                try:
                    if hasattr(data, 'get_slice'):
                        if type(cut_axis) is int:
                            cut_axis = data.dimension_labels[cut_axis]
                        temp_dict = {cut_axis:cut_slice}
                        plot_data = data.get_slice(**temp_dict, force=True)
                    elif hasattr(data,'as_array'):
                        plot_data = data.as_array().take(indices=cut_slice, axis=cut_axis)
                    else:
                        plot_data = data.take(indices=cut_slice, axis=cut_axis)
                except:
                    raise TypeError("Unable to slice input data. Could not obtain 2D slice {0} from {1} with shape {2}.\n\
                            Pass either correct slice information or a 2D array".format(slice_requested, type(data), data.shape))

                subtitle = "direction: {0}, slice: {1}".format(cut_axis,cut_slice)
            else:
                plot_data = data
                subtitle = None


            #check dataset is now 2D
            if len(plot_data.shape) != 2:
                raise TypeError("Unable to slice input data. Could not obtain 2D slice {0} from {1} with shape {2}.\n\
                                Pass either correct slice information or a 2D array".format(slice_requested, type(data), data.shape))

            #get axis labels per subplot
            if type(axis_labels) is list:
                plot_axis_labels = axis_labels[i]
            else:
                plot_axis_labels = axis_labels

            if plot_axis_labels is None and hasattr(plot_data,'dimension_labels'):
                    plot_axis_labels = (plot_data.dimension_labels[1],plot_data.dimension_labels[0])

            #get min/max of subsetted data
            range_min = min(range_min, plot_data.min())
            range_max = max(range_max, plot_data.max())

            #get title per subplot
            if isinstance(title, list):
                if title[i] is None:
                    plot_title = ''
                else:
                    plot_title = title[i]
            else:
                if title is None:
                    plot_title = ''
                else:
                    plot_title = title

            if subtitle is not None:
                plot_title += '\n' + subtitle

            #get origin per subplot
            if isinstance(origin, list):
                plot_origin = origin[i]
            else:
                plot_origin = origin

            subplots.append(_PlotData(plot_data,plot_title,plot_axis_labels, plot_origin))

        #set range per subplot
        for i, subplot in enumerate(subplots):
            if fix_range is False:
                pass
            elif fix_range is True:
                subplot.range = (range_min,range_max)
            elif type(fix_range) is list:
                subplot.range = fix_range[i]
            else:
                subplot.range = (fix_range[0], fix_range[1])

        #create plots
        if num_plots < num_cols:
            num_cols = num_plots

        num_rows = int(round((num_plots+0.5)/num_cols))
        fig, (ax) = plt.subplots(num_rows, num_cols, figsize=size)
        axes = ax.flatten()

        #set up plots
        for i in range(num_rows*num_cols):
            axes[i].set_visible(False)

        for i, subplot in enumerate(subplots):

            axes[i].set_visible(True)
            axes[i].set_title(subplot.title)

            if subplot.axis_labels is not None:
                axes[i].set_ylabel(subplot.axis_labels[1])
                axes[i].set_xlabel(subplot.axis_labels[0])

            #set origin
            data, data_origin, extent = set_origin(subplot.data, subplot.origin)
            if isinstance(cmap, (list, tuple)):
                dcmap = cmap[i]
            else:
                dcmap = cmap
            sp = axes[i].imshow(data, cmap=dcmap, origin=data_origin, extent=extent)

            im_ratio = subplot.data.shape[0]/subplot.data.shape[1]

            y_axes2 = False
            if isinstance(subplot.data,(AcquisitionData)):
                if axes[i].get_ylabel() == 'angle':
                    locs = axes[i].get_yticks()
                    location_new = locs[0:-1].astype(int)

                    ang = subplot.data.geometry.config.angles

                    labels_new = ["{:.2f}".format(i) for i in np.take(ang.angle_data, location_new)]
                    axes[i].set_yticks(location_new, labels=labels_new)

                    axes[i].set_ylabel('angle / ' + str(ang.angle_unit))

                    y_axes2 = axes[i].axes.secondary_yaxis('right')
                    y_axes2.set_ylabel('angle / index')

                    if subplot.data.shape[0] < subplot.data.shape[1]//2:
                        axes[i].set_aspect(1/im_ratio)
                        im_ratio = 1

            if y_axes2:
                scale = 0.041*im_ratio
                pad = 0.12
            else:
                scale = 0.0467*im_ratio
                pad = 0.02

            plt.colorbar(sp, orientation='vertical', ax=axes[i],fraction=scale, pad=pad)

            if subplot.range is not None:
                sp.set_clim(subplot.range[0],subplot.range[1])

        fig.set_tight_layout(True)
        fig.set_facecolor('w')

        #plt.show() creates a new figure so we save a copy to return
        fig2 = plt.gcf()
        plt.show()
        return fig2

def plotter2D(datacontainers, title=None, slice_list=None, fix_range=False, axis_labels=None, origin='lower-left', cmap='gray', num_cols=2, size=(15,15)):
    '''Alias of show2D'''
    return show2D(datacontainers, title=title, slice_list=slice_list, fix_range=fix_range, axis_labels=axis_labels, origin=origin, cmap=cmap, num_cols=num_cols, size=size)

class _Arrow3D(FancyArrowPatch):

    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)

    def do_3d_projection(self, renderer=None):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
        return np.min(zs)

class _ShowGeometry(object):
    def __init__(self, acquisition_geometry, image_geometry=None):
        if AcquisitionType.DIM2 & acquisition_geometry.dimension:
            self.ndim = 2
            sys = acquisition_geometry.config.system
            if acquisition_geometry.geom_type == 'cone':
                ag_temp = AcquisitionGeometry.create_Cone3D([*sys.source.position,0], [*sys.detector.position,0], [*sys.detector.direction_x,0],[0,0,1],[*sys.rotation_axis.position,0],[0,0,1])
            else:
                ag_temp = AcquisitionGeometry.create_Parallel3D([*sys.ray.direction,0], [*sys.detector.position,0], [*sys.detector.direction_x,0],[0,0,1],[*sys.rotation_axis.position,0],[0,0,1])


            ag_temp.config.panel = acquisition_geometry.config.panel
            ag_temp.set_angles(acquisition_geometry.angles)
            ag_temp.set_labels(['vertical', *acquisition_geometry.dimension_labels])

            self.acquisition_geometry = ag_temp

        elif acquisition_geometry.channels > 1:
            self.ndim = 3
            self.acquisition_geometry = acquisition_geometry.get_slice(channel=0)
        else:
            self.acquisition_geometry = acquisition_geometry
            self.ndim = 3

        if image_geometry is None:
            self.image_geometry=self.acquisition_geometry.get_ImageGeometry()
        else:
            self.image_geometry = image_geometry


        len1 = self.acquisition_geometry.config.panel.num_pixels[0] * self.acquisition_geometry.config.panel.pixel_size[0]
        len2 = self.acquisition_geometry.config.panel.num_pixels[1] * self.acquisition_geometry.config.panel.pixel_size[1]
        self.scale = max(len1,len2)/5


        self.handles = []
        self.labels = []

    def draw(self, elev=35, azim=35, view_distance=10, grid=False, figsize=(10,10), fontsize=10, show=True):

        self.fig = plt.figure(figsize=figsize)
        self.ax = self.fig.add_subplot(111, projection='3d')

        self.text_options = {   'horizontalalignment': 'center',
                                'verticalalignment': 'center',
                                'fontsize': fontsize }


        self.display_world()

        if self.acquisition_geometry.geom_type == 'cone':
            self.display_source()
        else:
            self.display_ray()

        self.display_object()

        self.display_detector()


        if grid is False:
            self.ax.set_axis_off()

        self.ax.view_init(elev=elev, azim=azim)
        self.ax.dist = view_distance

        #to force aspect ratio 1:1:1
        world_limits = self.ax.get_w_lims()
        self.ax.set_box_aspect((world_limits[1]-world_limits[0],world_limits[3]-world_limits[2],world_limits[5]-world_limits[4]))

        l = self.ax.plot(np.NaN, np.NaN, '-', color='none', label='')[0]

        for i in range(3):
            self.handles.insert(2,l)
            self.labels.insert(2,'')

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.ax.legend(self.handles, self.labels, loc='upper left', bbox_to_anchor= (0, 1), ncol=3,
                        borderaxespad=0, frameon=False,fontsize=self.text_options.get('fontsize'))


        self.fig.set_tight_layout(True)
        self.fig.set_facecolor('w')

        #plt.show() creates a new figure so we save a copy to return
        fig2 = plt.gcf()

        if show:
            plt.show()

        return fig2

    def display_world(self):

        self.ax.set_xlabel('X axis')
        self.ax.set_ylabel('Y axis')

        if self.ndim == 3:
            self.ax.set_zlabel('Z axis')
        else:
            self.ax.set_zticks([])

        #origin and coordinate frame
        Oo = np.zeros(3)
        self.ax.scatter3D(*Oo, marker='o', alpha=1,color='k',lw=1)
        h = mlines.Line2D([], [], color='k',linestyle='solid', markersize=12, label='world coordinate system')

        labels = ['$x$','$y$','$z$']
        for i in range(self.ndim):
            axis = np.zeros(3)
            axis[i] = 1 * self.scale

            a = _Arrow3D(*zip(Oo,axis*2), mutation_scale=20,lw=1, arrowstyle="->", color="k")
            self.ax.add_artist(a)

            self.ax.text(*(axis*2.2),labels[i], **self.text_options)

        self.handles.append(h)
        self.labels.append(h.get_label())

    def detector_vertex(self):
        # detector corners
        det_size  = (np.array(self.acquisition_geometry.config.panel.num_pixels) * np.array(self.acquisition_geometry.config.panel.pixel_size))/2
        det_rows_dir = self.acquisition_geometry.config.system.detector.direction_x

        if self.ndim == 3:
            det_v = self.acquisition_geometry.config.system.detector.direction_y * det_size[1]
            det_h = det_rows_dir * det_size[0]

            rt = det_h + det_v + self.acquisition_geometry.config.system.detector.position
            lt = -det_h + det_v + self.acquisition_geometry.config.system.detector.position
            lb = -det_h - det_v + self.acquisition_geometry.config.system.detector.position
            rb = det_h - det_v + self.acquisition_geometry.config.system.detector.position

            return [rb, lb, lt, rt]

        else:
            det_h = det_rows_dir * det_size[0]
            r = det_h  + self.acquisition_geometry.config.system.detector.position
            l = -det_h  + self.acquisition_geometry.config.system.detector.position
            return [r, l]


    def display_detector(self):

        do = self.acquisition_geometry.config.system.detector.position
        det = self.detector_vertex()

        #mark data origin
        if 'right' in self.acquisition_geometry.config.panel.origin:
            if self.ndim==2 or 'bottom' in self.acquisition_geometry.config.panel.origin:
                pix0 = det[0]
            else:
                pix0 = det[3]
        else:
            if self.ndim==2 or 'bottom' in self.acquisition_geometry.config.panel.origin:
                pix0 = det[1]
            else:
                pix0 = det[2]

        det_rows_dir = self.acquisition_geometry.config.system.detector.direction_x

        x = _Arrow3D(*zip(do, self.scale * det_rows_dir + do), mutation_scale=20,lw=1, arrowstyle="-|>", color="b")
        self.ax.add_artist(x)
        self.ax.text(*(1.2 * self.scale * det_rows_dir + do),r'$D_x$', **self.text_options)

        if self.ndim == 3:
            det_col_dir = self.acquisition_geometry.config.system.detector.direction_y
            y = _Arrow3D(*zip(do, self.scale * det_col_dir + do), mutation_scale=20,lw=1, arrowstyle="-|>", color="b")
            self.ax.add_artist(y)
            self.ax.text(*(1.2 * self.scale * det_col_dir + do),r'$D_y$', **self.text_options)

        handles=[
            self.ax.scatter3D(*do, marker='o', alpha=1,color='b',lw=1, label='detector position'),
            mlines.Line2D([], [], color='b',linestyle='solid', markersize=12, label='detector direction'),
            self.ax.plot3D(*zip(*det, det[0]), color='b',ls='dotted',alpha=1, label='detector')[0],
            self.ax.scatter3D(*pix0, marker='x', alpha=1,color='b',lw=1,s=50, label='data origin (pixel 0)'),
        ]

        for x in handles:
            self.handles.append(x)
            self.labels.append(x.get_label())

    def display_object(self):

        ro = self.acquisition_geometry.config.system.rotation_axis.position
        h0 = self.ax.scatter3D(*ro, marker='o', alpha=1,color='r',lw=1,label='rotation axis position')
        self.handles.append(h0)
        self.labels.append(h0.get_label())

        if self.ndim == 3:
            # rotate axis arrow
            r1 = ro +  self.acquisition_geometry.config.system.rotation_axis.direction * self.scale * 2
            arrow3 = _Arrow3D(*zip(ro,r1), mutation_scale=20,lw=1, arrowstyle="-|>", color="r")
            self.ax.add_artist(arrow3)

            a = self.acquisition_geometry.config.system.rotation_axis.direction


            # draw reco
            x = np.array([self.image_geometry.get_min_x(), self.image_geometry.get_max_x()])
            y = np.array([self.image_geometry.get_min_y(), self.image_geometry.get_max_y()])
            z = np.array([self.image_geometry.get_min_z(), self.image_geometry.get_max_z()])

            combos = [
                ((x[0],y[0],z[0]),(x[0],y[1],z[0])),
                ((x[0],y[1],z[0]),(x[1],y[1],z[0])),
                ((x[1],y[1],z[0]),(x[1],y[0],z[0])),
                ((x[1],y[0],z[0]),(x[0],y[0],z[0])),
                ((x[0],y[0],z[1]),(x[0],y[1],z[1])),
                ((x[0],y[1],z[1]),(x[1],y[1],z[1])),
                ((x[1],y[1],z[1]),(x[1],y[0],z[1])),
                ((x[1],y[0],z[1]),(x[0],y[0],z[1])),
                ((x[0],y[0],z[0]),(x[0],y[0],z[1])),
                ((x[0],y[1],z[0]),(x[0],y[1],z[1])),
                ((x[1],y[1],z[0]),(x[1],y[1],z[1])),
                ((x[1],y[0],z[0]),(x[1],y[0],z[1])),
            ]

            if np.allclose(a,[0,0,1]):
                axis_rotation = np.eye(3)
            elif np.allclose(a,[0,0,-1]):
                axis_rotation = np.eye(3)
                axis_rotation[1][1] = -1
                axis_rotation[2][2] = -1
            else:
                vx = np.array([[0, 0, -a[0]], [0, 0, -a[1]], [a[0], a[1], 0]])
                axis_rotation = np.eye(3) + vx + vx.dot(vx) *  1 / (1 + a[2])

            rotation_matrix = np.matrix.transpose(axis_rotation)

            count = 0
            for x in combos:
                s = rotation_matrix.dot(np.asarray(x[0]).reshape(3,1))
                e = rotation_matrix.dot(np.asarray(x[1]).reshape(3,1))

                x_data = float(s[0]) + ro[0], float(e[0]) + ro[0]
                y_data = float(s[1]) + ro[1], float(e[1]) + ro[1]
                z_data = float(s[2]) + ro[2], float(e[2]) + ro[2]

                self.ax.plot3D(x_data,y_data,z_data, color="r",ls='dotted',alpha=1)

                if count == 0:
                    vox0=(x_data[0],y_data[0],z_data[0])
                count+=1
        else:
            # draw square
            x = [self.image_geometry.get_min_x(), self.image_geometry.get_max_x()]
            y = [self.image_geometry.get_min_y(), self.image_geometry.get_max_y()]
            vertex = np.array([(x[0],y[0],0),(x[0],y[1],0),(x[1],y[1],0),(x[1],y[0],0)]) + ro
            self.ax.plot3D(*zip(*vertex, vertex[0]), color='r',ls='dotted',alpha=1)
            vox0=vertex[0]
            rotation_matrix = np.eye(3)

        #rotation direction
        points = 36
        x = [None]*points
        y = [None]*points
        z = [None]*points

        for i in range(points):
            theta = i * (np.pi * 1.8) /36
            point_i = np.array([np.sin(theta),-np.cos(theta),0]).reshape(3,1)
            point_rot = -self.scale*0.5*rotation_matrix.dot(point_i)

            x[i] = float(point_rot[0] + ro[0])
            y[i] = float(point_rot[1] + ro[1])
            z[i] = float(point_rot[2] + ro[2])

        self.ax.plot3D(x,y,z, color='r',ls="dashed",alpha=1)
        arrow4 = _Arrow3D(x[-2:],y[-2:],z[-2:],mutation_scale=20,lw=1, arrowstyle="-|>", color="r")
        self.ax.add_artist(arrow4)

        handles = [
            mlines.Line2D([], [], color='r',linestyle='solid', markersize=12, label='rotation axis direction'),
            mlines.Line2D([], [], color='r',linestyle='dotted', markersize=15, label='image geometry'),
            self.ax.scatter3D(*vox0, marker='x', alpha=1,color='r',lw=1,s=50, label='data origin (voxel 0)'),
            mlines.Line2D([], [], color='r',linestyle='dashed', markersize=12, label=r'rotation direction $\theta$')
        ]

        for x in handles:
            self.handles.append(x)
            self.labels.append(x.get_label())

    def display_source(self):

        so = self.acquisition_geometry.config.system.source.position
        det = self.detector_vertex()

        for i in range(len(det)):
            self.ax.plot3D(*zip(so,det[i]), color='#D4BD72',ls="dashed",alpha=0.4)

        self.ax.plot3D(*zip(so,self.acquisition_geometry.config.system.detector.position), color='#D4BD72',ls="solid",alpha=1)[0],

        h0 = self.ax.scatter3D(*so, marker='*', alpha=1,color='#D4BD72',lw=1, label='source position', s=100)

        self.handles.append(h0)
        self.labels.append(h0.get_label())

    def display_ray(self):

        det = self.detector_vertex()
        det.append(self.acquisition_geometry.config.system.detector.position)

        dist = np.sqrt(np.sum(self.acquisition_geometry.config.system.detector.position**2))*2

        if dist < 0.01:
            dist = self.acquisition_geometry.config.panel.num_pixels[0] * self.acquisition_geometry.config.panel.pixel_size[0]

        rays = det - self.acquisition_geometry.config.system.ray.direction*dist

        for i in range(len(rays)):
            h0 = self.ax.plot3D(*zip(rays[i],det[i]), color='#D4BD72',ls="dashed",alpha=0.4, label='ray direction')[0]
            arrow = _Arrow3D(*zip(rays[i],rays[i]+self.acquisition_geometry.config.system.ray.direction*self.scale ),mutation_scale=20,lw=1, arrowstyle="-|>", color="#D4BD72")
            self.ax.add_artist(arrow)

        self.handles.append(h0)
        self.labels.append(h0.get_label())

class show_geometry(show_base):
    '''
    Displays a schematic of the acquisition geometry
    for 2D geometries elevation and azimuthal cannot be changed


    Parameters
    ----------
    acquisition_geometry: AcquisitionGeometry
        CIL acquisition geometry
    image_geometry: ImageGeometry, optional
        CIL image geometry
    elevation: float
        Camera elevation in degrees, 3D geometries only, default=20
    azimuthal: float
        Camera azimuthal in degrees, 3D geometries only, default=-35
    view_distance: float
        Camera view distance, default=10
    grid: boolean
        Show figure axis, default=False
    figsize: tuple (x, y)
        Set figure size (inches), default (10,10)
    fontsize: int
        Set fontsize, default 10
    show: bool
        Show Matplotlib figure window, default=True
        
    Note
    ----

    The figure can be saved using the `save` method . i.e:

    >>> fig = show_geometry(geometry)
    >>> fig.save('/path/to/output.png')
    
    '''



    def __init__(self,acquisition_geometry, image_geometry=None, elevation=20, azimuthal=-35, view_distance=10, grid=False, figsize=(10,10), fontsize=10, show=True):
        if AcquisitionType.CONE_FLEX & acquisition_geometry.geom_type:
            raise NotImplementedError("The `cone_flex` geometry type is not supported by show_geometry. Use `show_system_positions` instead.")


        if AcquisitionType.DIM2 & acquisition_geometry.dimension:
            elevation = 90
            azimuthal = 0

        self.display = _ShowGeometry(acquisition_geometry, image_geometry)

        self.figure = self.display.draw(elev=elevation, azim=azimuthal, view_distance=view_distance, grid=grid, figsize=figsize, fontsize=fontsize, show=show)

class show_system_positions(show_base):
    '''
    Displays four plots to show i) the source position, 
    ii) the detector centre, iii) the detector x-direction, and 
    iv) the detector y-direction for each projection.


    Parameters
    ----------
    acquisition_geometry: AcquisitionGeometry
        CIL acquisition geometry
    figsize: tuple (x, y)
        Set figure size (inches), default (10,10)
    fontsize: int
        Set fontsize, default 10

    Returns
    -------
    matplotlib.figure.Figure
        returns a matplotlib.pyplot figure object
    '''
    def __init__(self, acquisition_geometry:AcquisitionGeometry, figsize=(10,10), fontsize=10):

        # Only applicable for AcquisitionGeometry
        if not isinstance(acquisition_geometry, AcquisitionGeometry):
            raise ValueError(f"The data type of `acquisition_geometry` must be \"<class 'cil.framework.AcquisitionGeometry'>\". It is \"{type(acquisition_geometry)}\", which is not currently supported by this function.")

        # Only applicable for cone_flex geometry type
        if acquisition_geometry.geom_type != AcquisitionType.CONE_FLEX:
            raise ValueError(f"The geometry type of `acquisition_geometry` must be \"cone_flex\". It is \"{acquisition_geometry.geom_type}\", which is not currently supported by this function.")

        self.figure = self._draw(acquisition_geometry, figsize, fontsize)

    def _draw(self, acquisition_geometry, figsize, fontsize):

        # Plot the data
        self.fig, self.axs = plt.subplots(2, 2, figsize=figsize)
        system  = acquisition_geometry.config.system

        x_axis_values = np.arange(acquisition_geometry.num_projections)
        i = 0; j = 0
        x = 0; y = 1; z = 2
        self.axs[j,i].set_title("Source position")

        x_pos = np.array([vec.position[0] for vec in system.source])
        y_pos = np.array([vec.position[1] for vec in system.source])
        z_pos = np.array([vec.position[2] for vec in system.source])

        self.axs[j,i].plot(x_axis_values, x_pos, label="X axis")
        self.axs[j,i].plot(x_axis_values, y_pos, label="Y axis")
        self.axs[j,i].plot(x_axis_values, z_pos, label="Z axis")
        self.axs[j,i].legend(fontsize=fontsize)
        # self.axs[j,i].set_xlabel("Projection #")
        self.axs[j,i].set_ylabel("Position")

        i = 1; j = 0
        x += 3; y += 3; z += 3
        self.axs[j,i].set_title("Detector Center")

        x_pos = np.array([vec.position[0] for vec in system.detector])
        y_pos = np.array([vec.position[1] for vec in system.detector])
        z_pos = np.array([vec.position[2] for vec in system.detector])

        self.axs[j,i].plot(x_axis_values, x_pos, label="X axis")
        self.axs[j,i].plot(x_axis_values, y_pos, label="Y axis")
        self.axs[j,i].plot(x_axis_values, z_pos, label="Z axis")
        self.axs[j,i].legend(fontsize=fontsize)
        # self.axs[j,i].set_xlabel("Projection #")
        # self.axs[j,i].set_ylabel("Position in (cm)")

        i = 0; j = 1
        x += 3; y += 3; z += 3
        self.axs[j,i].set_title("Detector X-direction")

        x_pos = np.array([vec.direction_x[0] for vec in system.detector])
        y_pos = np.array([vec.direction_x[1] for vec in system.detector])
        z_pos = np.array([vec.direction_x[2] for vec in system.detector])

        self.axs[j,i].plot(x_axis_values, x_pos, label="X axis")
        self.axs[j,i].plot(x_axis_values, y_pos, label="Y axis")
        self.axs[j,i].plot(x_axis_values, z_pos, label="Z axis")

        self.axs[j,i].legend(fontsize=fontsize)
        self.axs[j,i].set_xlabel("Projection #")
        self.axs[j,i].set_ylabel("Position")

        i = 1; j = 1
        x += 3; y += 3; z += 3
        self.axs[j,i].set_title("Detector Y-direction")

        x_pos = np.array([vec.direction_y[0] for vec in system.detector])
        y_pos = np.array([vec.direction_y[1] for vec in system.detector])
        z_pos = np.array([vec.direction_y[2] for vec in system.detector])

        self.axs[j,i].plot(x_axis_values, x_pos, label="X axis")
        self.axs[j,i].plot(x_axis_values, y_pos, label="Y axis")
        self.axs[j,i].plot(x_axis_values, z_pos, label="Z axis")
        self.axs[j,i].legend(fontsize=fontsize)
        self.axs[j,i].set_xlabel("Projection #")
        # self.axs[j,i].set_ylabel("Position in (mm)")

        # Resize the text
        for ax in self.axs.flatten():
            for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()):
                item.set_fontsize(fontsize)

        plt.tight_layout()
        fig2 = plt.gcf()
        return fig2




#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import (DataProcessor, AcquisitionData, ImageData, DataContainer, ImageGeometry, VectorGeometry,
                           AcquisitionGeometry)
from cil.framework.labels import AcquisitionType
import numpy as np
import weakref
import logging

log = logging.getLogger(__name__)


# Note to developers: Binner and Slicer share a lot of common code
# so Binner has been implemented as a child of Slicer.  This makes use
# of commonality and redefines only the methods that differ. These methods
# dictate the style of slicer
class Slicer(DataProcessor):

    """This creates a Slicer processor.

    The processor will crop the data, and then return every n input pixels along a dimension from the starting index.

    The output will be a data container with the data, and geometry updated to reflect the operation.

    Parameters
    ----------

    roi : dict
        The region-of-interest to slice {'axis_name1':(start,stop,step), 'axis_name2':(start,stop,step)}
        The `key` being the axis name to apply the processor to, the `value` holding a tuple containing the ROI description

        Start: Starting index of input data. Must be an integer, or `None` defaults to index 0.
        Stop: Stopping index of input data. Must be an integer, or `None` defaults to index N.
        Step: Number of pixels to average together. Must be an integer or `None` defaults to 1.


    Example
    -------

    >>> from cil.processors import Slicer
    >>> roi = {'horizontal':(10,-10,2),'vertical':(10,-10,2)}
    >>> processor = Slicer(roi)
    >>> processor.set_input(data)
    >>> data_sliced= processor.get_output()


    Example
    -------
    >>> from cil.processors import Slicer
    >>> roi = {'horizontal':(None,None,2),'vertical':(None,None,2)}
    >>> processor = Slicer(roi)
    >>> processor.set_input(data.geometry)
    >>> geometry_sliced = processor.get_output()


    Note
    ----
    The indices provided are start inclusive, stop exclusive.

    All elements along a dimension will be included if the axis does not appear in the roi dictionary, or if passed as {'axis_name',-1}

    If only one number is provided, then it is interpreted as Stop. i.e. {'axis_name1':(stop)}
    If two numbers are provided, then they are interpreted as Start and Stop  i.e. {'axis_name1':(start, stop)}

    Negative indexing can be used to specify the index. i.e. {'axis_name1':(10, -10)} will crop the dimension symmetrically

    If Stop - Start is not multiple of Step, then
    the resulted dimension will have (Stop - Start) // Step
    elements, i.e. (Stop - Start) % Step elements will be ignored

    """

    def __init__(self,
                 roi = None):

        kwargs = {
            '_roi_input': roi,
            '_roi_ordered':None,
            '_data_array': False,
            '_geometry': None,
            '_processed_dims':None,
            '_shape_in':None,
            '_shape_out_full':None,
            '_shape_out':None,
            '_labels_out':None,
            '_labels_in':None,
            '_pixel_indices':None,
            '_accelerated': True
            }

        super(Slicer, self).__init__(**kwargs)


    def set_input(self, dataset):
        """
        Set the input data or geometry to the processor

        Parameters
        ----------
        dataset : DataContainer, Geometry
            The input DataContainer or Geometry
        """

        if issubclass(type(dataset), DataContainer) or isinstance(dataset,(AcquisitionGeometry,ImageGeometry)):
            if self.check_input(dataset):
                self.__dict__['input'] = weakref.ref(dataset)
                self.__dict__['shouldRun'] = True
            else:
                raise ValueError('Input data not compatible')
        else:
            raise TypeError("Input type mismatch: got {0} expecting {1}"\
                            .format(type(dataset), DataContainer))

        self._set_up()

    def check_input(self, data):

        if isinstance(data, (ImageData,AcquisitionData)):
            self._data_array = True
            self._geometry = data.geometry

        elif isinstance(data, DataContainer):
            self._data_array = True
            self._geometry = None

        elif isinstance(data, (ImageGeometry, AcquisitionGeometry)):
            self._data_array = False
            self._geometry = data

        else:
            raise TypeError('Processor supports following data types:\n' +
                            ' - ImageData\n - AcquisitionData\n - DataContainer\n - ImageGeometry\n - AcquisitionGeometry')

        if self._data_array:
            if data.dtype != np.float32:
                raise TypeError("Expected float32")

        if (self._roi_input == None):
            raise ValueError('Please, specify roi')

        for key in self._roi_input.keys():
            if key not in data.dimension_labels:
                raise ValueError('Wrong label is specified for roi, expected one of {}.'.format(data.dimension_labels))
            if isinstance(self._geometry , (AcquisitionGeometry)) and self._geometry.geom_type & AcquisitionType.CONE_FLEX \
                    and key in ['vertical', 'horizontal']:
                raise NotImplementedError("Cone-Flex geometry is not supported by this processor for slicing along 'vertical' or 'horizontal'")


        return True


    def _set_up(self):
        """
        This parses the input roi generically and then configures the processor according to its class.
        """
        #read input
        data = self.get_input()
        self._parse_roi(data.ndim, data.shape, data.dimension_labels)
        #processor specific configurations
        self._configure()
        # set boolean of dimensions to process
        self._processed_dims = [0 if self._shape_out_full[i] == self._shape_in[i] else 1 for i in range(4)]
        self._shape_out = tuple([i for i in self._shape_out_full if i > 1])
        self._labels_out = [self._labels_in[i] for i,x in enumerate(self._shape_out_full) if x > 1]

    def _parse_roi(self, ndim, shape, dimension_labels):
        '''
        Process the input roi
        '''
        offset = 4-ndim
        labels_in = [None]*4
        labels_in[offset::] = dimension_labels
        shape_in = [1]*4
        shape_in[offset::] = shape

        # defaults
        range_list = [range(0,x, 1) for x in shape_in]

        for i in range(ndim):

            roi = self._roi_input.get(dimension_labels[i],None)

            if roi == None or roi == -1:
                continue

            start = range_list[offset + i].start
            stop = range_list[offset + i].stop
            step = range_list[offset + i].step

            # accepts a tuple, range or slice
            try:
                roi = [roi.start, roi.stop, roi.step]
            except AttributeError:
                roi = list(roi)

            length = len(roi)

            if length == 1:
                if roi[0] is not None:
                    stop = roi[0]
            elif length > 1:
                if roi[0] is not None:
                    start = roi[0]
                if roi[1] is not None:
                    stop = roi[1]

            if length > 2:
                if roi[2] is not None:
                    step = roi[2]

            # deal with negative indexing
            if start < 0:
                start += shape_in[offset + i]

            if stop <= 0:
                stop += shape_in[offset + i]

            if stop > shape_in[offset+i]:
                log.warning(f"ROI for axis {dimension_labels[i]} has 'stop' out of bounds. Using axis length as stop value."
                            f" Got stop index: {stop}, using {shape_in[offset+i]}")
                stop = shape_in[offset+i]

            if start > shape_in[offset+i]:
                raise ValueError(f"ROI for axis {dimension_labels[i]} has 'start' out of bounds."
                                 f" Got start index: {start} for axis length {shape_in[offset+i]}")

            if start >= stop:
                raise ValueError(f"ROI for axis {dimension_labels[i]} has 'start' out of bounds."
                                 f" Got start index: {start}, stop index {stop}")

            # set values
            range_list[offset+ i] = range(int(start), int(stop), int(step))

        # set values
        self._shape_in = shape_in
        self._labels_in = labels_in
        self._roi_ordered = range_list


    def _configure(self):
        """
        Once the ROI has been parsed this configure the input specifically for use with Slicer
        """
        self._shape_out_full = [len(x) for x in self._roi_ordered]
        self._pixel_indices = [(x[0],x[-1]) for x in self._roi_ordered]


    def _get_slice_position(self, roi):
        """
        Return the vertical position to extract a single slice for sliced geometry
        """
        return roi.start


    def _get_angles(self, roi):
        """
        Returns the sliced angles according to the roi
        """
        return self._geometry.angles[roi.start:roi.stop:roi.step]

    def _process_acquisition_geometry(self):
        """
        Creates the new acquisition geometry
        """
        geometry_new = self._geometry.copy()

        processed_dims = self._processed_dims.copy()

        # deal with vertical first as it may change the geometry type
        if 'vertical' in self._geometry.dimension_labels:
            vert_ind = self._labels_in.index('vertical')
            if processed_dims[vert_ind]:
                roi = self._roi_ordered[vert_ind]
                n_elements = len(roi)

                if n_elements > 1:
                    # difference in end indices, minus differences in start indices, divided by 2
                    pixel_offset = ((self._shape_in[vert_ind] -1 - self._pixel_indices[vert_ind][1]) - self._pixel_indices[vert_ind][0])*0.5
                    geometry_new.config.shift_detector_in_plane(pixel_offset, 'vertical')
                    geometry_new.config.panel.num_pixels[1] = n_elements
                else:
                    try:
                        position = self._get_slice_position(roi)
                        geometry_new = geometry_new.get_slice(vertical = position)
                    except ValueError:
                        log.warning("Unable to calculate the requested 2D geometry. Returning geometry=`None`")
                        return None

                geometry_new.config.panel.pixel_size[1] *= roi.step
                processed_dims[vert_ind] = False


        for i, axis  in enumerate(self._labels_in):

            if not processed_dims[i]:
                continue

            roi = self._roi_ordered[i]
            n_elements = len(roi)

            if axis == 'channel':
                geometry_new.set_channels(num_channels=n_elements)

            elif axis == 'angle':
                geometry_new.config.angles.angle_data = self._get_angles(roi)

            elif axis == 'projection':
                geometry_new.config.system.num_positions = int(np.ceil((roi.stop - roi.start )/ roi.step))
                geometry_new.config.system.source = self._geometry.config.system.source[roi.start:roi.stop:roi.step]
                geometry_new.config.system.detector = self._geometry.config.system.detector[roi.start:roi.stop:roi.step]
                

            elif axis == 'horizontal':
                pixel_offset = ((self._shape_in[i] -1 - self._pixel_indices[i][1]) - self._pixel_indices[i][0])*0.5

                geometry_new.config.shift_detector_in_plane(pixel_offset, axis)
                geometry_new.config.panel.num_pixels[0] = n_elements
                geometry_new.config.panel.pixel_size[0] *= roi.step

        return geometry_new


    def _process_image_geometry(self):
        """
        Creates the new image geometry
        """

        if len(self._shape_out) == 0:
            return None
        elif len(self._shape_out) ==1:
            return VectorGeometry(self._shape_out[0], dimension_labels=self._labels_out[0])

        geometry_new = self._geometry.copy()
        for i, axis in enumerate(self._labels_in):

            if not self._processed_dims[i]:
                continue

            roi = self._roi_ordered[i]
            n_elements = len(roi)

            voxel_offset = (self._shape_in[i] -1 - self._pixel_indices[i][1] - self._pixel_indices[i][0])*0.5

            if axis == 'channel':
                geometry_new.channels = n_elements
                geometry_new.channel_spacing *= roi.step

            elif axis == 'vertical':
                geometry_new.center_z -= voxel_offset * geometry_new.voxel_size_z

                geometry_new.voxel_num_z = n_elements
                geometry_new.voxel_size_z *= roi.step

            elif axis == 'horizontal_x':
                geometry_new.center_x -= voxel_offset * geometry_new.voxel_size_x

                geometry_new.voxel_num_x = n_elements
                geometry_new.voxel_size_x *= roi.step

            elif axis == 'horizontal_y':
                geometry_new.center_y -= voxel_offset * geometry_new.voxel_size_y

                geometry_new.voxel_num_y = n_elements
                geometry_new.voxel_size_y *= roi.step

        return geometry_new


    def _process_data(self, dc_in, dc_out):
        """
        Slice the data array
        """
        slice_obj = tuple([slice(x.start, x.stop, x.step) for x in self._roi_ordered])
        arr_in = dc_in.array.reshape(self._shape_in)
        dc_out.fill(np.squeeze(arr_in[slice_obj]))

    def process(self, out=None):
        """
        Processes the input data

        Parameters
        ----------
        out : ImageData, AcquisitionData, DataContainer, optional
           Fills the referenced DataContainer with the processed output and suppresses the return

        Returns
        -------
        DataContainer
            The downsampled output is returned. Depending on the input type this may be:
            ImageData, AcquisitionData, DataContainer, ImageGeometry, AcquisitionGeometry
        """
        data = self.get_input()

        if isinstance(self._geometry, ImageGeometry):
            new_geometry = self._process_image_geometry()
        elif isinstance(self._geometry, AcquisitionGeometry):
            new_geometry = self._process_acquisition_geometry()
        else:
            new_geometry = None

        print("New geometry: ", new_geometry)
        print("Shape out: ", self._shape_out)

        # return if just acting on geometry
        if not self._data_array:
            return new_geometry

        # create output array or check size and shape of passed out
        if out is None:
            if new_geometry is not None:
                data_out = new_geometry.allocate(None)
                print("New geometry shape: ", data_out.shape)
            else:
                processed_array = np.empty(self._shape_out,dtype=np.float32)
                data_out = DataContainer(processed_array,False, self._labels_out)
        else:
            try:
                out.array = np.asarray(out.array, dtype=np.float32, order='C').reshape(self._shape_out)
            except:
                raise ValueError("Array of `out` not compatible. Expected shape: {0}, data type: {1} Got shape: {2}, data type: {3}".format(self._shape_out, np.float32, out.array.shape, out.array.dtype))

            if new_geometry is not None:
                if out.geometry != new_geometry:
                    raise ValueError("Geometry of `out` not as expected. Got {0}, expected {1}".format(out.geometry, new_geometry))

            data_out = out


        self._process_data(data, data_out)


        return data_out



#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import ImageData
from cil.framework.labels import ImageDimension
import tomophantom
from tomophantom import TomoP2D, TomoP3D
import os
import numpy as np

import ctypes, platform
from ctypes import util
# check for the extension
if platform.system() == 'Linux':
    dll = 'libctomophantom.so'
elif platform.system() == 'Windows':
    dll_file = 'ctomophantom.dll'
    dll = util.find_library(dll_file)
elif platform.system() == 'Darwin':
    dll = 'libctomophantom.dylib'
else:
    raise ValueError('Not supported platform, ', platform.system())

libtomophantom = ctypes.cdll.LoadLibrary(dll)



path = os.path.dirname(tomophantom.__file__)
path_library2D = os.path.join(path, "Phantom2DLibrary.dat")
path_library3D = os.path.join(path, "Phantom3DLibrary.dat")

def is_model_temporal(num_model, num_dims=2):
    '''Returns whether a model in the TomoPhantom library is temporal

    This will go to check the installed library files from TomoPhantom
    https://github.com/dkazanc/TomoPhantom/tree/master/PhantomLibrary/models

    :param num_model: model number
    :type num_model: int
    :param num_dims: dimensionality of the phantom, 2D or 3D
    :type num_dims: int, default 2
    '''
    return get_model_num_channels(num_model, num_dims) > 1

def get_model_num_channels(num_model, num_dims=2):
    '''Returns number of temporal steps (channels) the model has

    This will go to check the installed library files from TomoPhantom
    https://github.com/dkazanc/TomoPhantom/tree/master/PhantomLibrary/models

    https://github.com/dkazanc/TomoPhantom/blob/v1.4.9/Core/utils.c#L27
    https://github.com/dkazanc/TomoPhantom/blob/v1.4.9/Core/utils.c#L269

    :param num_model: model number
    :type num_model: int
    :param num_dims: dimensionality of the phantom, 2D or 3D
    :type num_dims: int, default 2
    '''

    return check_model_params(num_model, num_dims=num_dims)[3]

def check_model_params(num_model, num_dims=2):
    '''Returns params_switch array from the C TomoPhantom library in function checkParams2D or checkParams3D

    This will go to check the installed library files from TomoPhantom
    https://github.com/dkazanc/TomoPhantom/tree/master/PhantomLibrary/models

    https://github.com/dkazanc/TomoPhantom/blob/v1.4.9/Core/utils.c#L27
    https://github.com/dkazanc/TomoPhantom/blob/v1.4.9/Core/utils.c#L269

    :param num_model: model number
    :type num_model: int
    :param num_dims: dimensionality of the phantom, 2D or 3D
    :type num_dims: int, default 2
    '''
    if num_dims == 2:
        libtomophantom.checkParams2D.argtypes = [ctypes.POINTER(ctypes.c_int),  # pointer to the params array
                                  ctypes.c_int,                                   # model number selector (int)
                                  ctypes.c_char_p]                  # string to the library file
        params = np.zeros([10], dtype=np.int32)
        params_p = params.ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        lib2d_p = str(path_library2D).encode('utf-8')
        libtomophantom.checkParams2D(params_p, num_model, lib2d_p)

        return params

    elif num_dims == 3:
        libtomophantom.checkParams3D.argtypes = [ctypes.POINTER(ctypes.c_int),  # pointer to the params array
                                  ctypes.c_int,                                   # model number selector (int)
                                  ctypes.c_char_p]                  # string to the library file
        params = np.zeros([11], dtype=np.int32)
        params_p = params.ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        lib2d_p = str(path_library3D).encode('utf-8')
        libtomophantom.checkParams3D(params_p, num_model, lib2d_p)

        return params

    else:
        raise ValueError('Unsupported dimensionality. Expected 2 or 3, got {}'.format(dims))

def get_ImageData(num_model, geometry):
    '''Returns an ImageData relative to geometry with the model num_model from tomophantom

    :param num_model: model number
    :type num_model: int
    :param geometry: geometrical info that describes the phantom
    :type geometry: ImageGeometry
    Example usage:

    .. code-block:: python

      ndim = 2
      N=128
      angles = np.linspace(0, 360, 50, True, dtype=np.float32)
      offset = 0.4
      channels = 3

      if ndim == 2:
          ag = AcquisitionGeometry.create_Cone2D((offset,-100), (offset,100))
          ag.set_panel(N)

      else:
          ag = AcquisitionGeometry.create_Cone3D((offset,-100, 0), (offset,100,0))
          ag.set_panel((N,N-2))

      ag.set_channels(channels)
      ag.set_angles(angles, angle_unit=AngleUnit.DEGREE)

      ig = ag.get_ImageGeometry()
      num_model = 1
      phantom = TomoPhantom.get_ImageData(num_model=num_model, geometry=ig)


    '''
    ig = geometry.copy()
    ig.set_labels(ImageDimension.get_order_for_engine('cil'))
    num_dims = len(ig.dimension_labels)

    if ImageDimension.CHANNEL in ig.dimension_labels:
        if not is_model_temporal(num_model):
            raise ValueError('Selected model {} is not a temporal model, please change your selection'.format(num_model))
        if num_dims == 4:
            # 3D+time for tomophantom
            # output dimensions channel and then spatial,
            # e.g. [ 'channel', 'vertical', 'horizontal_y', 'horizontal_x' ]
            num_model = num_model
            shape = tuple(ig.shape[1:])
            phantom_arr = TomoP3D.ModelTemporal(num_model, shape, path_library3D)
        elif num_dims == 3:
            # 2D+time for tomophantom
            # output dimensions channel and then spatial,
            # e.g. [ 'channel', 'horizontal_y', 'horizontal_x' ]
            N = ig.shape[1]
            num_model = num_model
            phantom_arr = TomoP2D.ModelTemporal(num_model, ig.shape[1], path_library2D)
        else:
            raise ValueError('Wrong ImageGeometry')
        if ig.channels != phantom_arr.shape[0]:
            raise ValueError('The required model {} has {} channels. The ImageGeometry you passed has {}. Please update your ImageGeometry.'\
                .format(num_model, ig.channels, phantom_arr.shape[0]))
    else:
        if num_dims == 3:
            # 3D
            num_model = num_model
            phantom_arr = TomoP3D.Model(num_model, ig.shape, path_library3D)
        elif num_dims == 2:
            # 2D
            if ig.shape[0] != ig.shape[1]:
                raise ValueError('Can only handle square ImageData, got shape'.format(ig.shape))
            N = ig.shape[0]
            num_model = num_model
            phantom_arr = TomoP2D.Model(num_model, N, path_library2D)
        else:
            raise ValueError('Wrong ImageGeometry')


    im_data = ImageData(phantom_arr, geometry=ig)
    im_data.reorder(list(geometry.dimension_labels))
    return im_data



#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import DataProcessor, AcquisitionData, ImageData, DataContainer, ImageGeometry
import warnings
import numpy
from scipy import special, ndimage

class MaskGenerator(DataProcessor):
    r'''
    Processor to detect outliers and return a mask with 0 where outliers were detected, and 1 for other pixels. Please use the desiried method to configure a processor for your needs.
    '''

    @staticmethod
    def special_values(nan=True, inf=True):
        r'''This creates a MaskGenerator processor which generates a mask for inf and/or nan values.

        :param nan: mask NaN values
        :type nan: bool, default=True
        :param inf: mask INF values
        :type inf: bool, default=True

        '''
        if nan is True:
            if inf is True:
                processor = MaskGenerator(mode='special_values')
            else:
                processor = MaskGenerator(mode='nan')
        else:
            if inf is True:
                processor = MaskGenerator(mode='inf')
            else:
                raise ValueError("Please specify at least one type of value to threshold on")

        return processor

    @staticmethod
    def threshold(min_val=None, max_val=None):
        r'''This creates a MaskGenerator processor which generates a mask for values outside boundaries

        :param min_val: lower boundary
        :type min_val: float, default=None
        :param max_val: upper boundary
        :type max_val: float, default=None
        '''
        processor = MaskGenerator(mode='threshold', threshold_value=(min_val,max_val))
        return processor

    @staticmethod
    def quantile(min_quantile=None, max_quantile=None):
        r'''This creates a MaskGenerator processor which generates a mask for values outside boundaries

        :param min_quantile: lower quantile, 0-1
        :type min_quantile: float, default=None
        :param max_quantile: upper quantile, 0-1
        :type max_quantile: float, default=None
        '''
        processor = MaskGenerator(mode='quantile', quantiles=(min_quantile,max_quantile))
        return processor

    @staticmethod
    def mean(axis=None, threshold_factor=3, window=None):
        r'''This creates a MaskGenerator processor which generates a mask for values outside a multiple of standard-devaiations from the mean.

        abs(A - mean(A)) < threshold_factor * std(A).

        :param threshold_factor: scale factor of standard-deviations to use as threshold
        :type threshold_factor: float, default=3
        :param axis: specify axis as int or from 'dimension_labels' to calculate mean. If no axis is specified then operates over flattened array.
        :type axis: int, string
        :param window: specify number of pixels to use in calculation of a rolling mean
        :type window: int, default=None
        '''
        if window == None:
            processor = MaskGenerator(mode='mean', threshold_factor=threshold_factor, axis=axis)
        else:
            processor = MaskGenerator(mode='movmean', threshold_factor=threshold_factor, axis=axis, window=window)

        return processor

    @staticmethod
    def median(axis=None, threshold_factor=3, window=None):
        r'''This creates a MaskGenerator processor which generates a mask for values outside a multiple of median absolute deviation (MAD) from the mean.

        abs(A - median(A)) < threshold_factor * MAD(A),
        MAD = c*median(abs(A-median(A))) where c=-1/(sqrt(2)*erfcinv(3/2))

        :param threshold_factor: scale factor of MAD to use as threshold
        :type threshold_factor: float, default=3
        :param axis: specify axis as int or from 'dimension_labels' to calculate mean. If no axis is specified then operates over flattened array.
        :type axis: int, string
        :param window: specify number of pixels to use in calculation of a rolling median
        :type window: int, default=None
        '''

        if window == None:
            processor = MaskGenerator(mode='median', threshold_factor=threshold_factor, axis=axis)
        else:
            processor = MaskGenerator(mode='movmedian', threshold_factor=threshold_factor, axis=axis, window=window)

        return processor

    def __init__(self,
                 mode='special_values',
                 threshold_value=(None, None),
                 quantiles=(None, None),
                 threshold_factor=3,
                 window=5,
                 axis=None):
        r'''Processor to detect outliers and return mask with 0 where outliers were detected and 1 for other pixels.

            :param mode: a method for detecting outliers (special_values, nan, inf, threshold, quantile, mean, median, movmean, movmedian)
            :type mode: string, default='special_values'
            :param threshold_value: specify lower and upper boundaries if 'threshold' mode is selected
            :type threshold_value: tuple
            :param quantiles: specify lower and upper quantiles if 'quantile' mode is selected
            :type quantiles: tuple
            :param threshold_factor: scales detection threshold (standard deviation in case of 'mean', 'movmean' and median absolute deviation in case of 'median', movmedian')
            :type threshold_factor: float, default=3
            :param window: specify running window if 'movmean' or 'movmedian' mode is selected
            :type window: int, default=5
            :param axis: specify axis to alculate statistics for 'mean', 'median', 'movmean', 'movmean' modes
            :type axis: int, string
            :return: returns a DataContainer with boolean mask with 0 where outliers were detected
            :rtype: DataContainer

        - special_values    test element-wise for both inf and nan
        - nan               test element-wise for nan
        - inf               test element-wise for inf
        - threshold         test element-wise if array values are within boundaries
                            given by threshold_values = (float,float).
                            You can secify only lower threshold value by setting another to None
                            such as threshold_values = (float,None), then
                            upper boundary will be amax(data). Similarly, to specify only upper
                            boundary, use threshold_values = (None,float). If both threshold_values
                            are set to None, then original array will be returned.
        - quantile          test element-wise if array values are within boundaries
                            given by quantiles = (q1,q2), 0<=q1,q2<=1.
                            You can secify only lower quantile value by setting another to None
                            such as quantiles = (float,q2), then
                            upper boundary will be amax(data). Similarly, to specify only upper
                            boundary, use quantiles = (None,q1). If both quantiles
                            are set to None, then original array will be returned.
        - mean              test element-wise if
                            abs(A - mean(A)) < threshold_factor * std(A).
                            Default value of threshold_factor is 3. If no axis is specified,
                            then operates over flattened array. Alternatively operates along axis specified
                            as dimension_label.
        - median            test element-wise if
                            abs(A - median(A)) < threshold_factor * scaled MAD(A),
                            scaled median absolute deviation (MAD) is defined as
                            c*median(abs(A-median(A))) where c=-1/(sqrt(2)*erfcinv(3/2))
                            Default value of threshold_factor is 3. If no axis is specified,
                            then operates over flattened array. Alternatively operates along axis specified
                            as dimension_label.
        - movmean           the same as mean but uses rolling mean with a specified window,
                            default window value is 5
        - movmedian         the same as mean but uses rolling median with a specified window,
                            default window value is 5

        '''

        kwargs = {'mode': mode,
                'threshold_value': threshold_value,
                'threshold_factor': threshold_factor,
                'quantiles': quantiles,
                'window': window,
                'axis': axis}

        super(MaskGenerator, self).__init__(**kwargs)

    def check_input(self, data):

        if self.mode not in ['special_values', 'nan', 'inf', 'threshold', 'quantile',
                             'mean', 'median', 'movmean', 'movmedian']:
            raise Exception("Wrong mode. One of the following is expected:\n" +
                            "special_values, nan, inf, threshold, \n quantile, mean, median, movmean, movmedian")

        if self.axis is not None and type(self.axis) is not int:
            if self.axis not in data.dimension_labels:
                raise Exception("Wrong label is specified for axis. " +
                                "Expected {}, got {}.".format(data.dimension_labels, self.axis))

        return True
    
    def check_output(self, out):
        if out is not None:
            if out.array.dtype != bool:
                raise TypeError("Input type mismatch: got {0} expecting {1}"\
                            .format(out.array.dtype, bool))
        
        return True


    def process(self, out=None):

        # get input DataContainer
        data = self.get_input()

        try:
            arr = data.as_array()
        except:
            arr = data

        ndim = arr.ndim

        try:
            axis_index = data.dimension_labels.index(self.axis)
        except:
            if type(self.axis) == int:
                axis_index = self.axis
            else:
                axis_index = None

        # intialise mask with all ones
        mask = numpy.ones(arr.shape, dtype=bool)

        # if NaN or +/-Inf
        if self.mode == 'special_values':

            mask[numpy.logical_or(numpy.isnan(arr), numpy.isinf(arr))] = 0

        elif self.mode == 'nan':

            mask[numpy.isnan(arr)] = 0

        elif self.mode == 'inf':

            mask[numpy.isinf(arr)] = 0

        elif self.mode == 'threshold':

            if not(isinstance(self.threshold_value, tuple)):
                raise Exception("Threshold value must be given as a tuple containing two values,\n" +\
                    "use None if no threshold value is given")

            threshold = self._parse_threshold_value(arr, quantile=False)

            mask[numpy.logical_or(arr < threshold[0], arr > threshold[1])] = 0

        elif self.mode == 'quantile':

            if not(isinstance(self.quantiles, tuple)):
                raise Exception("Quantiles must be given as a tuple containing two values,\n " + \
                    "use None if no quantile value is given")

            quantile = self._parse_threshold_value(arr, quantile=True)

            mask[numpy.logical_or(arr < quantile[0], arr > quantile[1])] = 0

        elif self.mode == 'mean':

            # if mean along specific axis
            if axis_index is not None:
                tile_par = []
                slice_obj = []
                for i in range(ndim):
                    if i == axis_index:
                        tile_par.append(axis_index)
                        slice_obj.append(numpy.newaxis)
                    else:
                        tile_par.append(1)
                        slice_obj.append(slice(None, None, 1))
                tile_par = tuple(tile_par)
                slice_obj = tuple(slice_obj)

                tmp_mean = numpy.tile((numpy.mean(arr, axis=axis_index))[slice_obj], tile_par)
                tmp_std = numpy.tile((numpy.std(arr, axis=axis_index))[slice_obj], tile_par)
                mask[numpy.abs(arr - tmp_mean) > self.threshold_factor * tmp_std] = 0

            # if global mean
            else:

                 mask[numpy.abs(arr - numpy.mean(arr)) > self.threshold_factor * numpy.std(arr)] = 0

        elif self.mode == 'median':

            c = -1 / (numpy.sqrt(2) * special.erfcinv(3 / 2))

            # if median along specific axis
            if axis_index is not None:
                tile_par = []
                slice_obj = []
                for i in range(ndim):
                    if i == axis_index:
                        tile_par.append(axis_index)
                        slice_obj.append(numpy.newaxis)
                    else:
                        tile_par.append(1)
                        slice_obj.append(slice(None, None, 1))
                tile_par = tuple(tile_par)
                slice_obj = tuple(slice_obj)

                tmp = numpy.abs(arr - numpy.tile((numpy.median(arr, axis=axis_index))[slice_obj], tile_par))
                median_absolute_dev = numpy.tile((numpy.median(tmp, axis=axis_index))[slice_obj], tile_par)
                mask[tmp > self.threshold_factor * c * median_absolute_dev] = 0

            # if global median
            else:

                tmp = numpy.abs(arr - numpy.median(arr))
                mask[tmp > self.threshold_factor * c * numpy.median(tmp)] = 0

        elif self.mode == 'movmean':

            # if movmean along specific axis
            if axis_index is not None:
                kernel = [1] * ndim
                kernel[axis_index] = self.window
                kernel = tuple(kernel)

                mean_array = ndimage.generic_filter(arr, numpy.mean, size=kernel, mode='reflect')
                std_array = ndimage.generic_filter(arr, numpy.std, size=kernel, mode='reflect')

                mask[numpy.abs(arr - mean_array) > self.threshold_factor * std_array] = 0

            # if global movmean
            else:
                mean_array = ndimage.generic_filter(arr, numpy.mean, size=(self.window,)*ndim, mode='reflect')
                std_array = ndimage.generic_filter(arr, numpy.std, size=(self.window,)*ndim, mode='reflect')

                mask[numpy.abs(arr - mean_array) > self.threshold_factor * std_array] = 0

        elif self.mode == 'movmedian':

            c = -1 / (numpy.sqrt(2) * special.erfcinv(3 / 2))

            # if movmedian along specific axis
            if axis_index is not None:

                # construct filter kernel
                kernel_shape = []
                for i in range(ndim):
                    if i == axis_index:
                        kernel_shape.append(self.window)
                    else:
                        kernel_shape.append(1)

                kernel_shape = tuple(kernel_shape)

                median_array = ndimage.median_filter(arr, footprint=kernel_shape, mode='reflect')

                tmp = abs(arr - median_array)
                mask[tmp > self.threshold_factor * c * ndimage.median_filter(tmp, footprint=kernel_shape, mode='reflect')] = 0

            # if global movmedian
            else:
                # construct filter kernel
                kernel_shape = tuple([self.window]*ndim)
                median_array = ndimage.median_filter(arr, size=kernel_shape, mode='reflect')

                tmp = abs(arr - median_array)
                mask[tmp > self.threshold_factor * c * ndimage.median_filter(tmp, size=kernel_shape, mode='reflect')] = 0

        else:
            raise ValueError('Mode not recognised. One of the following is expected: ' + \
                              'special_values, nan, inf, threshold, quantile, mean, median, movmean, movmedian')


        if out is None:
            mask = numpy.asarray(mask, dtype=bool)
            if data.geometry is not None:
                geometry = data.geometry.copy()
            else:
                geometry = None
            out = type(data)(mask, deep_copy=False, dtype=mask.dtype, geometry=geometry, dimension_labels=data.dimension_labels)
        else:
            out.fill(mask)
        
        return out

    def _parse_threshold_value(self, arr, quantile=False):

        lower_val = None
        upper_val = None

        if quantile == True:
            if self.quantiles[0] is not None:
                lower_val = numpy.quantile(arr, self.quantiles[0])
            if self.quantiles[1] is not None:
                upper_val = numpy.quantile(arr, self.quantiles[1])
        else:
            if self.threshold_value[0] is not None:
                lower_val = self.threshold_value[0]
            if self.threshold_value[1] is not None:
                upper_val = self.threshold_value[1]

        if lower_val is None:
            lower_val = numpy.amin(arr)

        if upper_val is None:
            upper_val = numpy.amax(arr)

        if upper_val <= lower_val:
            raise Exception("Upper threshold value must be larger than " + \
                "lower treshold value or min of data")

        return (lower_val, upper_val)
